# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An object to store the heuristic payoff table for a game."""

import abc
import collections

import numpy as np

from open_spiel.python.egt import utils


def _inc_average(count, average, value):
  """Computes the incremental average, `a_n = ((n - 1)a_{n-1} + v_n) / n`."""
  count += 1
  average = ((count - 1) * average + value) / count
  return (count, average)


def from_match_results(df, consider_agents):
  """Builds a heuristic payoff table from average win probabilities.

  Args:
    df: a Pandas dataframe of match results. Must contain a column "agents"
      consisting of tuples of agent names, and a column "scores" consisting of
      the score for each agent in the match.
    consider_agents: a list of agent names. Will only consider matches in which
      exclusively these agents appeared.

  Returns:
    A PayoffTable object.

  Raises:
    ValueError: if dataframe is empty, or columns 'agents' and 'scores' not
    specified, or games have zero players.
  """
  if df.empty:
    raise ValueError("Please provide a non-empty dataframe.")
  if "agents" not in df.columns:
    raise ValueError("Dataframe must contain a column 'agents'.")
  if "scores" not in df.columns:
    raise ValueError("Dataframe must contain a column 'scores'.")

  num_strategies = len(consider_agents)
  num_players = len(df["agents"][0])

  if num_players == 0:
    raise ValueError("Games must have > 0 players.")

  count_per_distribution = {}
  win_prob_per_distribution = {}

  for i, row in df.iterrows():
    print("Parsing row {} / {} ...".format(i, len(df)), end="\r")
    agents = row["agents"]
    scores = row["scores"]
    assert len(agents) == len(scores) == num_players

    if not set(agents).issubset(set(consider_agents)):
      # Ignore agents outside those we are supposed to consider.
      continue
    elif len(set(agents)) == 1:
      # Special case of self-play: deal with separately.
      continue

    # Find winner(s): In each match one must determine a winning strategy. One
    # way of doing this is to average over the returns for each strategy and
    # then say that the one with the greatest returns is the winner.

    # Get unique score per agent by averaging.
    count_per_agent = collections.defaultdict(int)
    average_score_per_agent = collections.defaultdict(int)
    for agent, score in zip(agents, scores):
      count_per_agent[agent], average_score_per_agent[agent] = _inc_average(
          count_per_agent[agent], average_score_per_agent[agent], score)

    winner_score = max(average_score_per_agent.values())
    winner_agents = [
        k for k, v in average_score_per_agent.items() if v == winner_score
    ]
    winner_strategy_idxs = [
        consider_agents.index(winner) for winner in winner_agents
    ]

    # Select the winner as the one maximizing the selected statistics.
    win_probabilities = np.zeros(num_strategies)
    for winner_strategy_idx in winner_strategy_idxs:
      win_probabilities[winner_strategy_idx] = 1 / len(winner_strategy_idxs)

    distribution = np.zeros(num_strategies)
    for agent, count in count_per_agent.items():
      strategy_idx = consider_agents.index(agent)
      distribution[strategy_idx] = count

    distribution = tuple(distribution)

    if distribution not in count_per_distribution:
      count_per_distribution[distribution] = 1
      win_prob_per_distribution[distribution] = win_probabilities
      continue
    (count_per_distribution[distribution],
     win_prob_per_distribution[distribution]) = _inc_average(
         count_per_distribution[distribution],
         win_prob_per_distribution[distribution], win_probabilities)

  # Populate self-play case (strategy both wins and loses).
  for idx, agent in enumerate(consider_agents):
    distribution = np.zeros(num_strategies)
    distribution[idx] = num_players
    distribution = tuple(distribution)
    win_prob = np.zeros(num_strategies)
    win_prob[idx] = 0.5
    win_prob_per_distribution[distribution] = win_prob

  # Create empty (nan) payoff table.
  table = PayoffTable(num_players, num_strategies)

  # Populate with win probabilities.
  for distribution, payoff in win_prob_per_distribution.items():
    table[distribution] = payoff

  return table


def from_matrix_game(matrix_game):
  """Returns a PayOffTable given a symmetric 2-player matrix game.

  Args:
    matrix_game: The payoff matrix corresponding to a 2-player symmetric game.
  """

  if not isinstance(matrix_game, np.ndarray):
    raise ValueError("The matrix game should be a numpy array, not a {}".format(
        type(matrix_game)))
  num_strats_per_population = (
      utils.get_num_strats_per_population(
          payoff_tables=[matrix_game], payoffs_are_hpt_format=False))
  assert len(num_strats_per_population) == 2
  assert num_strats_per_population[0] == num_strats_per_population[1]
  num_strategies = num_strats_per_population[0]

  num_profiles = utils.get_num_profiles(num_strats_per_population)
  table = PayoffTable(num_players=2, num_strategies=num_strategies)

  # Construct the HPT by filling in the corresponding payoffs for each profile
  for id_profile in range(num_profiles):
    strat_profile = utils.get_strat_profile_from_id(num_strats_per_population,
                                                    id_profile)
    distribution = table.get_distribution_from_profile(strat_profile)
    # For symmetric matrix games, multiple strategy profiles correspond to the
    # same distribution and payoffs. Thus, ensure the table entry has not
    # already been filled by a previous strategy profile.
    if table.item_is_uninitialized(tuple(distribution)):
      payoffs = np.zeros(num_strategies)
      payoffs[strat_profile[0]] = matrix_game[strat_profile[0],
                                              strat_profile[1]]
      payoffs[strat_profile[1]] = matrix_game[strat_profile[1],
                                              strat_profile[0]]
      table[tuple(distribution)] = payoffs

  return table


def from_heuristic_payoff_table(hpt):
  """Returns a `PayoffTable` instance from a numpy 2D HPT."""
  [num_rows, num_columns] = hpt.shape
  assert num_columns % 2 == 0
  num_strategies = int(num_columns / 2)
  num_players = np.sum(hpt[0, :num_strategies])
  obj = PayoffTable(num_players, num_strategies, initialize_payoff_table=False)

  # pylint: disable=protected-access
  for row in hpt:
    payoff_row = np.array(row[num_strategies:])
    obj._payoff_table[tuple(row[:num_strategies])] = payoff_row

  assert len(obj._payoff_table) == num_rows
  # pylint: enable=protected-access
  return obj


def _compute_win_probability_from_elo(rating_1, rating_2):
  """Computes the win probability of 1 vs 2 based on the provided Elo ratings.

  Args:
    rating_1: The Elo rating of player 1.
    rating_2: The Elo rating of player 2.

  Returns:
    The win probability of player 1, when playing against 2.
  """
  m = max(rating_1, rating_2)  # We subtract the max for numerical stability.

  m1 = 10**((rating_1 - m) / 400)
  m2 = 10**((rating_2 - m) / 400)

  return m1 / (m1 + m2)


def from_elo_scores(elo_ratings, num_agents=2):
  """Computes the Elo win probability payoff matrix `X` from the Elo scores.

  Args:
    elo_ratings: The elo scores vector of length [num_strategies].
    num_agents: The number of agents. Only 2 agents are supported for now.

  Returns:
    The HPT associated to the Elo win probability payoff matrix `X`. The score
    for a given agent is given by its win probability given its Elo score.

  Raises:
    ValueError: If `num_agents != 2`.
  """
  if num_agents != 2:
    raise ValueError("Only 2 agents are supported, because we need to compute "
                     "the win probability and that can only be computed with "
                     "2 players.")
  num_strategies = len(elo_ratings)

  hpt_rows = []

  possible_teams = utils.distribute(num_agents, num_strategies, normalize=False)

  for distribution_row in possible_teams:
    payoff_row = np.zeros([num_strategies])
    non_zero_index = np.nonzero(distribution_row)[0]  # Why [0]?
    assert len(non_zero_index.shape) == 1

    if len(non_zero_index) > 1:
      index_first_player, index_second_player = non_zero_index
      prob = _compute_win_probability_from_elo(elo_ratings[index_first_player],
                                               elo_ratings[index_second_player])
      payoff_row[index_first_player] = prob
      payoff_row[index_second_player] = 1 - prob
    elif len(non_zero_index) == 1:
      payoff_row[non_zero_index[0]] = 0.5
    else:
      assert False, "Impossible case, we have at least one strategy used."

    hpt_rows.append(np.hstack([distribution_row, payoff_row]))

  return NumpyPayoffTable(np.vstack(hpt_rows))


class _PayoffTableInterface(metaclass=abc.ABCMeta):
  """An interface for the PayoffTable classes."""

  @abc.abstractmethod
  def __call__(self):
    """Returns a view of the table as a np.array."""

  @abc.abstractproperty
  def num_strategies(self):
    pass

  @abc.abstractproperty
  def num_players(self):
    pass

  @abc.abstractproperty
  def num_rows(self):
    pass

  def expected_payoff(self, strategy):
    """The expected payoff of each pure strategy against the mixed strategy.

    We define the expected payoff of a strategy A as the expected payoff of
    that strategy over the space of 2 randomly sampled

    The mixed strategy is equivalently the composition of an infinitely large
    population. To find the expected payoff, we:
    1. Compute the probabilities of sampling each player distribution in the
       heuristic payoff table from the population.
    2. Compute the expected payoff of pure strategy against the mixed
       strategy by averaging over the payoff rows with these probabilities.

    For each pure strategy we must normalize by the probability that it appeared
    in the player distribution at all; otherwise we would be undercounting.

    For more details, see https://arxiv.org/pdf/1803.06376.pdf.

    Args:
      strategy: an `np.array(shape=self._num_strategies)` of probabilities.

    Returns:
      An `np.array(shape=self._num_strategies)` of payoffs for pure strategies.

    Raises:
      ValueError: if the provided strategy probabilities do not define a valid
        distribution over `self._num_strategies` strategies.
    """
    if strategy.shape != (self.num_strategies,):
      raise ValueError("The strategy probabilities should be of shape "
                       "({},), not {}".format(self.num_strategies,
                                              strategy.shape))
    if np.around(np.sum(strategy), decimals=3) != 1.0:
      raise ValueError("The strategy probabilities should sum to 1.")
    if not all([p >= 0 for p in strategy]):
      raise ValueError("The strategy probabilities should all be >= 0.")

    distributions = self._distributions.astype(int)
    if not np.all(np.isclose(self._distributions, distributions, 1e-10)):
      raise ValueError("Conversion to integers for distributions failed.")

    # Multinomial coefficients (one per distribution).
    coefficients = _multinomial_coefficients(distributions)
    # Probabilities of sampling each distribution given population composition.
    probabilities = _row_probabilities(coefficients, distributions, strategy)

    return _expected_payoff(probabilities, self._payoffs, strategy,
                            self._num_players)

  @property
  def _payoffs(self):
    """Returns an np.array containing the payoffs."""
    return self()[:, self.num_strategies:]

  @property
  def _distributions(self):
    """Returns an np.array containing the distribution over pure strategies."""
    return self()[:, :self.num_strategies]


class NumpyPayoffTable(object):
  """An object wrapping a Numpy array heuristic payoff table for a metagame.

  NOTE: We assume the number of players to be equal to the number of
  replicators.

  """

  def __init__(self, payoff_table, writeable=False):
    """Initializes an immutable payoff table.

    Let p be the number of players, k be the number of strategies. Then, there
    are Combinations(p + k - 1, k - 1) distinct configurations for the
    strategies of the p players.

    The payoff table is of shape [(p + k - 1)! / (p! * (k - 1)!), 2 * k].

    The first k columns encode the number of players playing each strategies.

    The second k columns encode the average payoff of each strategy in that
    game.

    Args:
      payoff_table: A numpy heuristic payoff table, which is assumed to be
        correctly constructed.
      writeable: Whether the numpy array payoff_table should be writeable. See
        https://docs.scipy.org/doc/numpy-1.15.1/reference/generated/numpy.ndarray.flags.html.
          However, locking a base object does not lock any views that already
          reference it,
    """
    self._writeable = writeable
    self._payoff_table = payoff_table

    [self._num_rows, num_columns] = self._payoff_table.shape
    assert num_columns % 2 == 0
    self._num_strategies = int(num_columns / 2)
    self._num_players = np.sum(self._payoff_table[0, :self._num_strategies])

  def __call__(self):
    """Returns a view of the table as a np.array.

    The mutability of the object is controlled by `writeable`.
    """
    if self._writeable:
      return self._payoff_table
    else:
      return np.copy(self._payoff_table)

  @property
  def writeable(self):
    return self._writeable

  @writeable.setter
  def writeable(self, writeable):
    self._writeable = writeable

  @property
  def num_strategies(self):
    return self._num_strategies

  @property
  def num_players(self):
    return self._num_players

  @property
  def num_rows(self):
    return self._num_rows


class PayoffTable(_PayoffTableInterface):
  """A mutable object to store the heuristic payoff table for a metagame."""

  def __init__(self, num_players, num_strategies, initialize_payoff_table=True):
    """A heuristic payoff table encodes payoffs from various strategy profiles.

    See `NumpyPayoffTable` for the description of the heuristic payoff table.

    Internally, this is represented as an OrderedDict {distribution: payoff}.

    Args:
      num_players: The number of players in the game.
      num_strategies: The number of strategies an individual could play.
      initialize_payoff_table: If `True`, nan entries will be created for all
        rows. If `False`, no rows are created at all.
    """
    super(PayoffTable, self).__init__()
    self.is_hpt = True
    self._num_players = num_players
    self._num_strategies = num_strategies
    self._payoff_table = collections.OrderedDict()

    if initialize_payoff_table:
      # Populate empty (nan) payoff table.
      player_distributions = utils.distribute(self._num_players,
                                              self._num_strategies)
      for d in player_distributions:
        self._payoff_table[d] = np.full(self._num_strategies, np.nan)

  def __call__(self):
    """Returns a view of the table as a np.array."""
    return np.concatenate((self._distributions, self._payoffs), axis=1)

  @property
  def _payoffs(self):
    """Returns an np.array containing the payoffs."""
    return np.array(list(self._payoff_table.values()))

  @property
  def _distributions(self):
    """Returns an np.array containing the distribution over pure strategies."""
    return np.array(list(self._payoff_table))

  @property
  def num_strategies(self):
    return self._num_strategies

  @property
  def num_players(self):
    return self._num_players

  @property
  def num_rows(self):
    return len(self._payoff_table)

  def __setitem__(self, distribution, payoff):
    assert distribution in self._payoff_table
    assert len(payoff) == self._num_strategies
    self._payoff_table[distribution] = payoff

  def __getitem__(self, distribution):
    """Returns the payoff profile for a given strategy distribution.

    Args:
      distribution: strategy profile tuple.

    Returns:
      Payoff profile for the corresponding strategy distribution.
    """
    return self._payoff_table[distribution]

  def item_is_uninitialized(self, distribution):
    return np.isnan(np.sum(self._payoff_table[distribution]))

  def get_distribution_from_profile(self, strat_profile):
    distribution = [0] * self.num_strategies
    for s in strat_profile:
      distribution[s] += 1
    return distribution


# The following provides utility functions to compute the expected payoff of
# a given strategy profile.
# See https://arxiv.org/pdf/1803.06376.pdf, page 3, left column.
#
# Usage:
#
# coefficients = _multinomial_coefficients(distributions, strategies):
# row_probabilities = _row_probabilities(coefficients, distributions, strategy)
# expected_payoff = _expected_payoff(row_probabilities, payoffs, composition,
#                                    num_players)
#
#
def _multinomial_coefficients(distributions):
  """Returns the multinomial coefficients.

  Args:
    distributions: The distributions table [num_rows, num_strategies].
  """
  v_factorial = np.vectorize(np.math.factorial)
  # Multinomial coefficients (one per distribution Ni).
  # (         P         )
  # ( Ni1, Ni1, ... Nik )
  coefficients = (
      v_factorial(np.sum(distributions, axis=1)) /
      np.prod(v_factorial(distributions), axis=1))

  return coefficients


def _row_probabilities(coefficients, distributions, strategy):
  """Returns the row probabilities [num_rows].

  Args:
    coefficients: The multinomial coefficients [num_rows].
    distributions: The distributions table [num_rows, num_strategies].
    strategy: The strategy array [num_strategies].
  """
  row_probabilities = coefficients * np.prod(
      np.power(strategy, distributions), axis=1)
  return row_probabilities


def _expected_payoff(row_probabilities, payoffs, strategy, num_players):
  # pylint: disable=g-doc-args
  r"""Returns the expected payoff.

  Computes (with p=num_players):

  r_j = \sum_i row_probabilities[i] * payoffs[i, j] / (1 - (1-strategy[j])^p)
  """
  # pylint: enable=g-doc-args
  [num_rows] = row_probabilities.shape
  [num_rows_2, num_strategies] = payoffs.shape
  [num_strategies_2] = strategy.shape
  assert num_rows == num_rows_2
  assert num_strategies == num_strategies_2

  # One per pure strategy.
  numerators = np.dot(np.transpose(payoffs), row_probabilities)
  # One per pure strategy.
  denominators = 1 - np.power(1 - strategy, num_players)
  return numerators / denominators



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Starting point for playing with the AlphaZero algorithm."""

from absl import app
from absl import flags

from open_spiel.python.algorithms.alpha_zero import alpha_zero
from open_spiel.python.algorithms.alpha_zero import model as model_lib
from open_spiel.python.utils import spawn

flags.DEFINE_string("game", "chess", "Name of the game.")
flags.DEFINE_integer("uct_c", 2, "UCT's exploration constant.")
flags.DEFINE_integer("max_simulations", 300, "How many simulations to run.")
flags.DEFINE_integer("train_batch_size", 2 ** 10, "Batch size for learning.")
flags.DEFINE_integer("replay_buffer_size", 2 ** 16,
                     "How many states to store in the replay buffer.")
flags.DEFINE_integer("replay_buffer_reuse", 3,
                     "How many times to learn from each state.")
flags.DEFINE_float("learning_rate", 0.001, "Learning rate.")
flags.DEFINE_float("weight_decay", 0.0001, "L2 regularization strength.")
flags.DEFINE_float("policy_epsilon", 0.25, "What noise epsilon to use.")
flags.DEFINE_float("policy_alpha", 1, "What dirichlet noise alpha to use.")
flags.DEFINE_float("temperature", 1,
                   "Temperature for final move selection.")
flags.DEFINE_integer("temperature_drop", 10,  # Less than AZ due to short games.
                     "Drop the temperature to 0 after this many moves.")
flags.DEFINE_enum("nn_model", "resnet", model_lib.Model.valid_model_types,
                  "What type of model should be used?.")
flags.DEFINE_integer("nn_width", 2 ** 7, "How wide should the network be.")
flags.DEFINE_integer("nn_depth", 10, "How deep should the network be.")
flags.DEFINE_string("path", None, "Where to save checkpoints.")
flags.DEFINE_integer("checkpoint_freq", 100, "Save a checkpoint every N steps.")
flags.DEFINE_integer("actors", 2, "How many actors to run.")
flags.DEFINE_integer("evaluators", 1, "How many evaluators to run.")
flags.DEFINE_integer("evaluation_window", 100,
                     "How many games to average results over.")
flags.DEFINE_integer(
    "eval_levels", 7,
    ("Play evaluation games vs MCTS+Solver, with max_simulations*10^(n/2)"
     " simulations for n in range(eval_levels). Default of 7 means "
     "running mcts with up to 1000 times more simulations."))
flags.DEFINE_integer("max_steps", 0, "How many learn steps before exiting.")
flags.DEFINE_bool("quiet", True, "Don't show the moves as they're played.")
flags.DEFINE_bool("verbose", False, "Show the MCTS stats of possible moves.")

FLAGS = flags.FLAGS


def main(unused_argv):
  config = alpha_zero.Config(
      game=FLAGS.game,
      path=FLAGS.path,
      learning_rate=FLAGS.learning_rate,
      weight_decay=FLAGS.weight_decay,
      train_batch_size=FLAGS.train_batch_size,
      replay_buffer_size=FLAGS.replay_buffer_size,
      replay_buffer_reuse=FLAGS.replay_buffer_reuse,
      max_steps=FLAGS.max_steps,
      checkpoint_freq=FLAGS.checkpoint_freq,

      actors=FLAGS.actors,
      evaluators=FLAGS.evaluators,
      uct_c=FLAGS.uct_c,
      max_simulations=FLAGS.max_simulations,
      policy_alpha=FLAGS.policy_alpha,
      policy_epsilon=FLAGS.policy_epsilon,
      temperature=FLAGS.temperature,
      temperature_drop=FLAGS.temperature_drop,
      evaluation_window=FLAGS.evaluation_window,
      eval_levels=FLAGS.eval_levels,

      nn_model=FLAGS.nn_model,
      nn_width=FLAGS.nn_width,
      nn_depth=FLAGS.nn_depth,
      observation_shape=None,
      output_size=None,

      quiet=FLAGS.quiet,
  )
  alpha_zero.alpha_zero(config)


if __name__ == "__main__":
  with spawn.main_handler():
    app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset generation for meta-CFR algorithm."""

from typing import List, Tuple

import numpy as np

from open_spiel.python.examples.meta_cfr.sequential_games.typing import InfostateNode


class Dataset:
  """Dataset class to generate data for training meta-CFR model."""

  def __init__(self, train_dataset: List[Tuple[List[List[float]],
                                               InfostateNode]],
               batch_size: int):
    self._train_dataset = np.array(train_dataset, dtype=object)
    self._size = self._train_dataset.shape[0]
    self._batch_size = batch_size

  def get_batch(self):
    while True:
      np.random.shuffle(self._train_dataset)
      idx_sample = np.random.choice(self._size, self._batch_size)
      next_batch = self._train_dataset[idx_sample, :]
      yield next_batch



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for softmax_policy."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import policy
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.algorithms import softmax_policy
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
import pyspiel


class SoftmaxPolicyTest(parameterized.TestCase):

  @parameterized.named_parameters(('python', 'python_mfg_crowd_modelling'),
                                  ('cpp', 'mfg_crowd_modelling'))
  def test_softmax(self, name):
    """Check if the softmax policy works as expected.

    The test checks that:
    - uniform prior policy gives the same results than no prior.
    - very high temperature gives almost a uniform policy.
    - very low temperature gives almost a deterministic policy for the best
    action.

    Args:
      name: Name of the game.
    """

    game = pyspiel.load_game(name)
    uniform_policy = policy.UniformRandomPolicy(game)
    dist = distribution.DistributionPolicy(game, uniform_policy)
    br_value = best_response_value.BestResponse(
        game, dist, value.TabularValueFunction(game))
    br_init_val = br_value(game.new_initial_state())

    # uniform prior policy gives the same results than no prior.
    softmax_pi_uniform_prior = softmax_policy.SoftmaxPolicy(
        game, None, 1.0, br_value, uniform_policy).to_tabular()
    softmax_pi_uniform_prior_value = policy_value.PolicyValue(
        game, dist, softmax_pi_uniform_prior, value.TabularValueFunction(game))
    softmax_pi_uniform_prior_init_val = softmax_pi_uniform_prior_value(
        game.new_initial_state())
    softmax_pi_no_prior = softmax_policy.SoftmaxPolicy(game, None, 1.0,
                                                       br_value, None)
    softmax_pi_no_prior_value = policy_value.PolicyValue(
        game, dist, softmax_pi_no_prior, value.TabularValueFunction(game))
    softmax_pi_no_prior_init_val = softmax_pi_no_prior_value(
        game.new_initial_state())

    self.assertAlmostEqual(softmax_pi_uniform_prior_init_val,
                           softmax_pi_no_prior_init_val)

    # very high temperature gives almost a uniform policy.
    uniform_policy = uniform_policy.to_tabular()
    uniform_value = policy_value.PolicyValue(game, dist, uniform_policy,
                                             value.TabularValueFunction(game))
    uniform_init_val = uniform_value(game.new_initial_state())

    softmax_pi_no_prior = softmax_policy.SoftmaxPolicy(game, None, 100000000,
                                                       br_value, None)
    softmax_pi_no_prior_value = policy_value.PolicyValue(
        game, dist, softmax_pi_no_prior, value.TabularValueFunction(game))
    softmax_pi_no_prior_init_val = softmax_pi_no_prior_value(
        game.new_initial_state())

    self.assertAlmostEqual(uniform_init_val, softmax_pi_no_prior_init_val)

    # very low temperature gives almost a best response policy.
    softmax_pi_no_prior = softmax_policy.SoftmaxPolicy(game, None, 0.0001,
                                                       br_value, None)
    softmax_pi_no_prior_value = policy_value.PolicyValue(
        game, dist, softmax_pi_no_prior, value.TabularValueFunction(game))
    softmax_pi_no_prior_init_val = softmax_pi_no_prior_value(
        game.new_initial_state())

    self.assertAlmostEqual(br_init_val, softmax_pi_no_prior_init_val)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Population Exploitability Descent (PED) Stochastic Approx. Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import updates


class Solver(updates.Solver):
  """PED Solver."""

  def __init__(self, proj_grad=True, euclidean=False, lrs=(1e-1,),
               rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    super().__init__(proj_grad, euclidean, rnd_init, seed)
    self.lrs = lrs

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see ped.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
      unregularized exploitability (stochastic estimate)
      unregularized exploitability (stochastic estimate) *duplicate
    """
    return gradients(*params, payoff_matrices, self.num_players, self.proj_grad)


def gradients(dist, payoff_matrices, num_players, proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    unregularized exploitability (stochastic estimate) *duplicate
  """
  nabla = payoff_matrices[0].dot(dist)

  power = np.inf
  s = np.linalg.norm(nabla, ord=power)
  br = np.zeros_like(dist)
  maxima = (nabla == s)
  br[maxima] = 1. / maxima.sum()

  unreg_exp = np.max(nabla) - nabla.dot(dist)

  grad_dist = -(nabla) + (num_players - 1) * payoff_matrices[1].dot(br - dist)
  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)

  return (grad_dist,), unreg_exp, unreg_exp



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests that Mean Field Games are implemented properly.

These tests are intended to help developers to write mean field games that
satisfy most of the unspecified constraints assumed by the following algorithms:
- python/mfg/algorithms/policy_value.py
- python/mfg/algorithms/nash_conv.py
- python/mfg/algorithms/mirror_descent.py
- python/mfg/algorithms/fictitious_play.py
- python/mfg/algorithms/distribution.py
- python/mfg/algorithms/best_response_value.py
- python/rl_environment.py
These tests are not exhaustive and will be updated with time.
"""
import random

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.mfg import games as mfg_games  # pylint:disable=unused-import
from open_spiel.python.mfg.algorithms import distribution
import pyspiel

FLAGS = flags.FLAGS

# Use a small depth limit to keep the length of the test reasonable.
flags.DEFINE_integer(
    'get_all_states_depth_limit', 10,
    'Depth limit of getting all the states (-1 for unlimited)')
flags.DEFINE_integer('rl_env_simulations', 10,
                     'Number of simulations for the RL environment tests')


def _get_next_states(state, next_states, to_string):
  """Extract non-chance states for a subgame into the all_states dict."""
  is_mean_field = state.current_player() == pyspiel.PlayerId.MEAN_FIELD
  if state.is_chance_node():
    # Add only if not already present

    for action, _ in state.chance_outcomes():
      next_state = state.child(action)
      state_str = to_string(next_state)
      if state_str not in next_states:
        next_states[state_str] = next_state

  if is_mean_field:
    support = state.distribution_support()
    next_state = state.clone()
    support_length = len(support)
    # update with a dummy distribution
    next_state.update_distribution(
        [1.0 / support_length for _ in range(support_length)])
    state_str = to_string(next_state)
    if state_str not in next_states:
      next_states[state_str] = next_state

  if int(state.current_player()) >= 0:
    for action in state.legal_actions():
      next_state = state.child(action)
      state_str = to_string(next_state)
      if state_str not in next_states:
        next_states[state_str] = next_state


def _next_states(states, to_string):
  next_states = {}
  for state in states:
    _get_next_states(state, next_states, to_string)
  return set(next_states.keys()), set(next_states.values())


def type_from_states(states):
  """Get node type of a list of states and assert they are the same."""
  types = [state.get_type() for state in states]
  assert len(set(types)) == 1
  return types[0]


class FiniteHorizonTest(parameterized.TestCase):

  @parameterized.parameters(
      {'game_name': 'python_mfg_crowd_modelling'},
      {'game_name': 'mfg_crowd_modelling'},
      {'game_name': 'mfg_garnet'},
      {'game_name': 'mfg_crowd_modelling_2d'},
      {'game_name': 'python_mfg_periodic_aversion'},
      {'game_name': 'python_mfg_predator_prey'},
  )
  def test_is_finite_horizon(self, game_name):
    """Check that the game has no loop."""
    game = pyspiel.load_game(game_name)
    states = set(game.new_initial_states())
    def to_string(s):
      return s.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    all_states_key = set(to_string(state) for state in states)
    while type_from_states(states) != pyspiel.StateType.TERMINAL:
      new_states_key, states = _next_states(states, to_string)
      self.assertEmpty(all_states_key.intersection(new_states_key))
      all_states_key.update(new_states_key)

  @parameterized.parameters(
      {'game_name': 'python_mfg_crowd_modelling'},
      {'game_name': 'mfg_crowd_modelling'},
      {'game_name': 'mfg_garnet'},
      {'game_name': 'mfg_crowd_modelling_2d'},
      {'game_name': 'python_mfg_periodic_aversion'},
      {'game_name': 'python_mfg_predator_prey'},
  )
  def test_has_at_least_an_action(self, game_name):
    """Check that all population's state have at least one action."""
    game = pyspiel.load_game(game_name)
    def to_string(s):
      return s.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    states = get_all_states.get_all_states(
        game,
        depth_limit=FLAGS.get_all_states_depth_limit,
        include_terminals=False,
        include_chance_states=False,
        include_mean_field_states=False,
        to_string=to_string)
    for state in states.values():
      self.assertNotEmpty(state.legal_actions())

  @parameterized.parameters(
      {'game_name': 'python_mfg_crowd_modelling'},
      {'game_name': 'mfg_crowd_modelling'},
      {'game_name': 'mfg_garnet'},
      {'game_name': 'mfg_crowd_modelling_2d'},
      {'game_name': 'python_mfg_periodic_aversion'},
      {'game_name': 'python_mfg_predator_prey'},
  )
  def test_rl_environment(self, game_name):
    """Check that the RL environment runs for a few trajectories."""
    game = pyspiel.load_game(game_name)
    uniform_policy = policy.UniformRandomPolicy(game)
    mfg_dist = distribution.DistributionPolicy(game, uniform_policy)

    envs = [
        rl_environment.Environment(
            game, mfg_distribution=mfg_dist, mfg_population=p)
        for p in range(game.num_players())
    ]
    for p, env in enumerate(envs):
      for _ in range(FLAGS.rl_env_simulations):
        time_step = env.reset()
        while not time_step.last():
          a = random.choice(time_step.observations['legal_actions'][p])
          time_step = env.step([a])

    env = envs[0]
    self.assertEqual(env.mfg_distribution, mfg_dist)

    # Update the distribution.
    new_mfg_dist = distribution.DistributionPolicy(game, uniform_policy)
    env.update_mfg_distribution(new_mfg_dist)
    self.assertEqual(env.mfg_distribution, new_mfg_dist)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example running PSRO on OpenSpiel Sequential games.

To reproduce results from (Muller et al., "A Generalized Training Approach for
Multiagent Learning", ICLR 2020; https://arxiv.org/abs/1909.12823), run this
script with:
  - `game_name` in ['kuhn_poker', 'leduc_poker']
  - `n_players` in [2, 3, 4, 5]
  - `meta_strategy_method` in ['alpharank', 'uniform', 'nash', 'prd']
  - `rectifier` in ['', 'rectified']

The other parameters keeping their default values.
"""

import time

from absl import app
from absl import flags
import numpy as np

# pylint: disable=g-bad-import-order
import pyspiel
import tensorflow.compat.v1 as tf
# pylint: enable=g-bad-import-order

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.algorithms import policy_aggregator
from open_spiel.python.algorithms.psro_v2 import best_response_oracle
from open_spiel.python.algorithms.psro_v2 import psro_v2
from open_spiel.python.algorithms.psro_v2 import rl_oracle
from open_spiel.python.algorithms.psro_v2 import rl_policy
from open_spiel.python.algorithms.psro_v2 import strategy_selectors


FLAGS = flags.FLAGS

# Game-related
flags.DEFINE_string("game_name", "kuhn_poker", "Game name.")
flags.DEFINE_integer("n_players", 2, "The number of players.")

# PSRO related
flags.DEFINE_string("meta_strategy_method", "alpharank",
                    "Name of meta strategy computation method.")
flags.DEFINE_integer("number_policies_selected", 1,
                     "Number of new strategies trained at each PSRO iteration.")
flags.DEFINE_integer("sims_per_entry", 1000,
                     ("Number of simulations to run to estimate each element"
                      "of the game outcome matrix."))

flags.DEFINE_integer("gpsro_iterations", 100,
                     "Number of training steps for GPSRO.")
flags.DEFINE_bool("symmetric_game", False, "Whether to consider the current "
                  "game as a symmetric game.")

# Rectify options
flags.DEFINE_string("rectifier", "",
                    "Which rectifier to use. Choices are '' "
                    "(No filtering), 'rectified' for rectified.")
flags.DEFINE_string("training_strategy_selector", "probabilistic",
                    "Which strategy selector to use. Choices are "
                    " - 'top_k_probabilities': select top "
                    "`number_policies_selected` strategies. "
                    " - 'probabilistic': Randomly samples "
                    "`number_policies_selected` strategies with probability "
                    "equal to their selection probabilities. "
                    " - 'uniform': Uniformly sample `number_policies_selected` "
                    "strategies. "
                    " - 'rectified': Select every non-zero-selection-"
                    "probability strategy available to each player.")

# General (RL) agent parameters
flags.DEFINE_string("oracle_type", "BR", "Choices are DQN, PG (Policy "
                    "Gradient) or BR (exact Best Response)")
flags.DEFINE_integer("number_training_episodes", int(1e4), "Number training "
                     "episodes per RL policy. Used for PG and DQN")
flags.DEFINE_float("self_play_proportion", 0.0, "Self play proportion")
flags.DEFINE_integer("hidden_layer_size", 256, "Hidden layer size")
flags.DEFINE_integer("batch_size", 32, "Batch size")
flags.DEFINE_float("sigma", 0.0, "Policy copy noise (Gaussian Dropout term).")
flags.DEFINE_string("optimizer_str", "adam", "'adam' or 'sgd'")

# Policy Gradient Oracle related
flags.DEFINE_string("loss_str", "qpg", "Name of loss used for BR training.")
flags.DEFINE_integer("num_q_before_pi", 8, "# critic updates before Pi update")
flags.DEFINE_integer("n_hidden_layers", 4, "# of hidden layers")
flags.DEFINE_float("entropy_cost", 0.001, "Self play proportion")
flags.DEFINE_float("critic_learning_rate", 1e-2, "Critic learning rate")
flags.DEFINE_float("pi_learning_rate", 1e-3, "Policy learning rate.")

# DQN
flags.DEFINE_float("dqn_learning_rate", 1e-2, "DQN learning rate.")
flags.DEFINE_integer("update_target_network_every", 1000, "Update target "
                     "network every [X] steps")
flags.DEFINE_integer("learn_every", 10, "Learn every [X] steps.")

# General
flags.DEFINE_integer("seed", 1, "Seed.")
flags.DEFINE_bool("local_launch", False, "Launch locally or not.")
flags.DEFINE_bool("verbose", True, "Enables verbose printing and profiling.")


def init_pg_responder(sess, env):
  """Initializes the Policy Gradient-based responder and agents."""
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  agent_class = rl_policy.PGPolicy

  agent_kwargs = {
      "session": sess,
      "info_state_size": info_state_size,
      "num_actions": num_actions,
      "loss_str": FLAGS.loss_str,
      "loss_class": False,
      "hidden_layers_sizes": [FLAGS.hidden_layer_size] * FLAGS.n_hidden_layers,
      "batch_size": FLAGS.batch_size,
      "entropy_cost": FLAGS.entropy_cost,
      "critic_learning_rate": FLAGS.critic_learning_rate,
      "pi_learning_rate": FLAGS.pi_learning_rate,
      "num_critic_before_pi": FLAGS.num_q_before_pi,
      "optimizer_str": FLAGS.optimizer_str
  }
  oracle = rl_oracle.RLOracle(
      env,
      agent_class,
      agent_kwargs,
      number_training_episodes=FLAGS.number_training_episodes,
      self_play_proportion=FLAGS.self_play_proportion,
      sigma=FLAGS.sigma)

  agents = [
      agent_class(  # pylint: disable=g-complex-comprehension
          env,
          player_id,
          **agent_kwargs)
      for player_id in range(FLAGS.n_players)
  ]
  for agent in agents:
    agent.freeze()
  return oracle, agents


def init_br_responder(env):
  """Initializes the tabular best-response based responder and agents."""
  random_policy = policy.TabularPolicy(env.game)
  oracle = best_response_oracle.BestResponseOracle(
      game=env.game, policy=random_policy)
  agents = [random_policy.__copy__() for _ in range(FLAGS.n_players)]
  return oracle, agents


def init_dqn_responder(sess, env):
  """Initializes the Policy Gradient-based responder and agents."""
  state_representation_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  agent_class = rl_policy.DQNPolicy
  agent_kwargs = {
      "session": sess,
      "state_representation_size": state_representation_size,
      "num_actions": num_actions,
      "hidden_layers_sizes": [FLAGS.hidden_layer_size] * FLAGS.n_hidden_layers,
      "batch_size": FLAGS.batch_size,
      "learning_rate": FLAGS.dqn_learning_rate,
      "update_target_network_every": FLAGS.update_target_network_every,
      "learn_every": FLAGS.learn_every,
      "optimizer_str": FLAGS.optimizer_str
  }
  oracle = rl_oracle.RLOracle(
      env,
      agent_class,
      agent_kwargs,
      number_training_episodes=FLAGS.number_training_episodes,
      self_play_proportion=FLAGS.self_play_proportion,
      sigma=FLAGS.sigma)

  agents = [
      agent_class(  # pylint: disable=g-complex-comprehension
          env,
          player_id,
          **agent_kwargs)
      for player_id in range(FLAGS.n_players)
  ]
  for agent in agents:
    agent.freeze()
  return oracle, agents


def print_policy_analysis(policies, game, verbose=False):
  """Function printing policy diversity within game's known policies.

  Warning : only works with deterministic policies.
  Args:
    policies: List of list of policies (One list per game player)
    game: OpenSpiel game object.
    verbose: Whether to print policy diversity information. (True : print)

  Returns:
    List of list of unique policies (One list per player)
  """
  states_dict = get_all_states.get_all_states(game, np.infty, False, False)
  unique_policies = []
  for player in range(len(policies)):
    cur_policies = policies[player]
    cur_set = set()
    for pol in cur_policies:
      cur_str = ""
      for state_str in states_dict:
        if states_dict[state_str].current_player() == player:
          pol_action_dict = pol(states_dict[state_str])
          max_prob = max(list(pol_action_dict.values()))
          max_prob_actions = [
              a for a in pol_action_dict if pol_action_dict[a] == max_prob
          ]
          cur_str += "__" + state_str
          for a in max_prob_actions:
            cur_str += "-" + str(a)
      cur_set.add(cur_str)
    unique_policies.append(cur_set)
  if verbose:
    print("\n=====================================\nPolicy Diversity :")
    for player, cur_set in enumerate(unique_policies):
      print("Player {} : {} unique policies.".format(player, len(cur_set)))
  print("")
  return unique_policies


def gpsro_looper(env, oracle, agents):
  """Initializes and executes the GPSRO training loop."""
  sample_from_marginals = True  # TODO(somidshafiei) set False for alpharank
  training_strategy_selector = FLAGS.training_strategy_selector or strategy_selectors.probabilistic

  g_psro_solver = psro_v2.PSROSolver(
      env.game,
      oracle,
      initial_policies=agents,
      training_strategy_selector=training_strategy_selector,
      rectifier=FLAGS.rectifier,
      sims_per_entry=FLAGS.sims_per_entry,
      number_policies_selected=FLAGS.number_policies_selected,
      meta_strategy_method=FLAGS.meta_strategy_method,
      prd_iterations=50000,
      prd_gamma=1e-10,
      sample_from_marginals=sample_from_marginals,
      symmetric_game=FLAGS.symmetric_game)

  start_time = time.time()
  for gpsro_iteration in range(FLAGS.gpsro_iterations):
    if FLAGS.verbose:
      print("Iteration : {}".format(gpsro_iteration))
      print("Time so far: {}".format(time.time() - start_time))
    g_psro_solver.iteration()
    meta_game = g_psro_solver.get_meta_game()
    meta_probabilities = g_psro_solver.get_meta_strategies()
    policies = g_psro_solver.get_policies()

    if FLAGS.verbose:
      print("Meta game : {}".format(meta_game))
      print("Probabilities : {}".format(meta_probabilities))

    # The following lines only work for sequential games for the moment.
    if env.game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL:
      aggregator = policy_aggregator.PolicyAggregator(env.game)
      aggr_policies = aggregator.aggregate(
          range(FLAGS.n_players), policies, meta_probabilities)

      exploitabilities, expl_per_player = exploitability.nash_conv(
          env.game, aggr_policies, return_only_nash_conv=False)

      _ = print_policy_analysis(policies, env.game, FLAGS.verbose)
      if FLAGS.verbose:
        print("Exploitabilities : {}".format(exploitabilities))
        print("Exploitabilities per player : {}".format(expl_per_player))


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  np.random.seed(FLAGS.seed)

  game = pyspiel.load_game_as_turn_based(FLAGS.game_name,
                                         {"players": FLAGS.n_players})
  env = rl_environment.Environment(game)

  # Initialize oracle and agents
  with tf.Session() as sess:
    if FLAGS.oracle_type == "DQN":
      oracle, agents = init_dqn_responder(sess, env)
    elif FLAGS.oracle_type == "PG":
      oracle, agents = init_pg_responder(sess, env)
    elif FLAGS.oracle_type == "BR":
      oracle, agents = init_br_responder(env)
    sess.run(tf.global_variables_initializer())
    gpsro_looper(env, oracle, agents)

if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Kuhn Poker implemented in Python.

This is a simple demonstration of implementing a game in Python, featuring
chance and imperfect information.

Python games are significantly slower than C++, but it may still be suitable
for prototyping or for small games.

It is possible to run C++ algorithms on Python implemented games, This is likely
to have good performance if the algorithm simply extracts a game tree and then
works with that. It is likely to be poor if the algorithm relies on processing
and updating states as it goes, e.g. MCTS.
"""

import enum

import numpy as np

import pyspiel


class Action(enum.IntEnum):
  PASS = 0
  BET = 1


_NUM_PLAYERS = 2
_DECK = frozenset([0, 1, 2])
_GAME_TYPE = pyspiel.GameType(
    short_name="python_kuhn_poker",
    long_name="Python Kuhn Poker",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=True,
    provides_observation_string=True,
    provides_observation_tensor=True,
    provides_factored_observation_string=True)
_GAME_INFO = pyspiel.GameInfo(
    num_distinct_actions=len(Action),
    max_chance_outcomes=len(_DECK),
    num_players=_NUM_PLAYERS,
    min_utility=-2.0,
    max_utility=2.0,
    utility_sum=0.0,
    max_game_length=3)  # e.g. Pass, Bet, Bet


class KuhnPokerGame(pyspiel.Game):
  """A Python version of Kuhn poker."""

  def __init__(self, params=None):
    super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return KuhnPokerState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    return KuhnPokerObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
        params)


class KuhnPokerState(pyspiel.State):
  """A python version of the Kuhn poker state."""

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self.cards = []
    self.bets = []
    self.pot = [1.0, 1.0]
    self._game_over = False
    self._next_player = 0

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every sequential-move game with chance.

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self._game_over:
      return pyspiel.PlayerId.TERMINAL
    elif len(self.cards) < _NUM_PLAYERS:
      return pyspiel.PlayerId.CHANCE
    else:
      return self._next_player

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    assert player >= 0
    return [Action.PASS, Action.BET]

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self.is_chance_node()
    outcomes = sorted(_DECK - set(self.cards))
    p = 1.0 / len(outcomes)
    return [(o, p) for o in outcomes]

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self.is_chance_node():
      self.cards.append(action)
    else:
      self.bets.append(action)
      if action == Action.BET:
        self.pot[self._next_player] += 1
      self._next_player = 1 - self._next_player
      if ((min(self.pot) == 2) or
          (len(self.bets) == 2 and action == Action.PASS) or
          (len(self.bets) == 3)):
        self._game_over = True

  def _action_to_string(self, player, action):
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      return f"Deal:{action}"
    elif action == Action.PASS:
      return "Pass"
    else:
      return "Bet"

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._game_over

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    pot = self.pot
    winnings = float(min(pot))
    if not self._game_over:
      return [0., 0.]
    elif pot[0] > pot[1]:
      return [winnings, -winnings]
    elif pot[0] < pot[1]:
      return [-winnings, winnings]
    elif self.cards[0] > self.cards[1]:
      return [winnings, -winnings]
    else:
      return [-winnings, winnings]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    return "".join([str(c) for c in self.cards] + ["pb"[b] for b in self.bets])


class KuhnPokerObserver:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, iig_obs_type, params):
    """Initializes an empty observation tensor."""
    if params:
      raise ValueError(f"Observation parameters not supported; passed {params}")

    # Determine which observation pieces we want to include.
    pieces = [("player", 2, (2,))]
    if iig_obs_type.private_info == pyspiel.PrivateInfoType.SINGLE_PLAYER:
      pieces.append(("private_card", 3, (3,)))
    if iig_obs_type.public_info:
      if iig_obs_type.perfect_recall:
        pieces.append(("betting", 6, (3, 2)))
      else:
        pieces.append(("pot_contribution", 2, (2,)))

    # Build the single flat tensor.
    total_size = sum(size for name, size, shape in pieces)
    self.tensor = np.zeros(total_size, np.float32)

    # Build the named & reshaped views of the bits of the flat tensor.
    self.dict = {}
    index = 0
    for name, size, shape in pieces:
      self.dict[name] = self.tensor[index:index + size].reshape(shape)
      index += size

  def set_from(self, state, player):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    self.tensor.fill(0)
    if "player" in self.dict:
      self.dict["player"][player] = 1
    if "private_card" in self.dict and len(state.cards) > player:
      self.dict["private_card"][state.cards[player]] = 1
    if "pot_contribution" in self.dict:
      self.dict["pot_contribution"][:] = state.pot
    if "betting" in self.dict:
      for turn, action in enumerate(state.bets):
        self.dict["betting"][turn, action] = 1

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    pieces = []
    if "player" in self.dict:
      pieces.append(f"p{player}")
    if "private_card" in self.dict and len(state.cards) > player:
      pieces.append(f"card:{state.cards[player]}")
    if "pot_contribution" in self.dict:
      pieces.append(f"pot[{int(state.pot[0])} {int(state.pot[1])}]")
    if "betting" in self.dict and state.bets:
      pieces.append("".join("pb"[b] for b in state.bets))
    return " ".join(str(p) for p in pieces)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, KuhnPokerGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.cfr."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import external_sampling_mccfr
import pyspiel

SEED = 39823987


class ExternalSamplingMCCFRTest(absltest.TestCase):

  def test_external_sampling_leduc_2p_simple(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("leduc_poker")
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.SIMPLE)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Leduc2P, conv = {}".format(conv))
    self.assertLess(conv, 5)
    # ensure that to_tabular() works on the returned policy and
    # the tabular policy is equivalent
    tabular_policy = es_solver.average_policy().to_tabular()
    conv2 = exploitability.nash_conv(game, tabular_policy)
    self.assertEqual(conv, conv2)

  def test_external_sampling_leduc_2p_full(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("leduc_poker")
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.FULL)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Leduc2P, conv = {}".format(conv))
    self.assertLess(conv, 5)

  def test_external_sampling_kuhn_2p_simple(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker")
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.SIMPLE)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Kuhn2P, conv = {}".format(conv))
    self.assertLess(conv, 1)

  def test_external_sampling_kuhn_2p_full(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker")
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.FULL)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Kuhn2P, conv = {}".format(conv))
    self.assertLess(conv, 1)

  # Liar's dice takes too long, so disable this test. Leave code for reference.
  # pylint: disable=g-unreachable-test-method
  def disabled_test_external_sampling_liars_dice_2p_simple(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("liars_dice")
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.SIMPLE)
    for _ in range(1):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Liar's dice, conv = {}".format(conv))
    self.assertLess(conv, 2)

  def test_external_sampling_kuhn_3p_simple(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker", {"players": 3})
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.SIMPLE)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Kuhn3P, conv = {}".format(conv))
    self.assertLess(conv, 2)

  def test_external_sampling_kuhn_3p_full(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker", {"players": 3})
    es_solver = external_sampling_mccfr.ExternalSamplingSolver(
        game, external_sampling_mccfr.AverageType.FULL)
    for _ in range(10):
      es_solver.iteration()
    conv = exploitability.nash_conv(game, es_solver.average_policy())
    print("Kuhn3P, conv = {}".format(conv))
    self.assertLess(conv, 2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Policy gradient agents trained and evaluated on Kuhn Poker."""

from absl import app
from absl import flags
from absl import logging
import tensorflow.compat.v1 as tf

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import policy_gradient

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", int(1e6), "Number of train episodes.")
flags.DEFINE_integer("eval_every", int(1e4), "Eval agents every x episodes.")
flags.DEFINE_enum("loss_str", "rpg", ["a2c", "rpg", "qpg", "rm"],
                  "PG loss to use.")


class PolicyGradientPolicies(policy.Policy):
  """Joint policy to be evaluated."""

  def __init__(self, env, nfsp_policies):
    game = env.game
    player_ids = [0, 1]
    super(PolicyGradientPolicies, self).__init__(game, player_ids)
    self._policies = nfsp_policies
    self._obs = {"info_state": [None, None], "legal_actions": [None, None]}

  def action_probabilities(self, state, player_id=None):
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)

    self._obs["current_player"] = cur_player
    self._obs["info_state"][cur_player] = (
        state.information_state_tensor(cur_player))
    self._obs["legal_actions"][cur_player] = legal_actions

    info_state = rl_environment.TimeStep(
        observations=self._obs, rewards=None, discounts=None, step_type=None)

    p = self._policies[cur_player].step(info_state, is_evaluation=True).probs
    prob_dict = {action: p[action] for action in legal_actions}
    return prob_dict


def main(_):
  game = "kuhn_poker"
  num_players = 2

  env_configs = {"players": num_players}
  env = rl_environment.Environment(game, **env_configs)
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  with tf.Session() as sess:
    # pylint: disable=g-complex-comprehension
    agents = [
        policy_gradient.PolicyGradient(
            sess,
            idx,
            info_state_size,
            num_actions,
            loss_str=FLAGS.loss_str,
            hidden_layers_sizes=(128,)) for idx in range(num_players)
    ]
    expl_policies_avg = PolicyGradientPolicies(env, agents)

    sess.run(tf.global_variables_initializer())
    for ep in range(FLAGS.num_episodes):

      if (ep + 1) % FLAGS.eval_every == 0:
        losses = [agent.loss for agent in agents]
        expl = exploitability.exploitability(env.game, expl_policies_avg)
        msg = "-" * 80 + "\n"
        msg += "{}: {}\n{}\n".format(ep + 1, expl, losses)
        logging.info("%s", msg)

      time_step = env.reset()
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        agent_output = agents[player_id].step(time_step)
        action_list = [agent_output.action]
        time_step = env.step(action_list)

      # Episode is over, step all agents with final info state.
      for agent in agents:
        agent.step(time_step)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the C++ nfg_writer methods exposed to Python."""

from absl.testing import absltest

import pyspiel


class NFGWriterTest(absltest.TestCase):

  def test_rps(self):
    expected_rps_nfg = ("""NFG 1 R "OpenSpiel export of matrix_rps()"
{ "Player 0" "Player 1" } { 3 3 }

0 0
1 -1
-1 1
-1 1
0 0
1 -1
1 -1
-1 1
0 0
""")
    game = pyspiel.load_game("matrix_rps")
    nfg_text = pyspiel.game_to_nfg_string(game)
    self.assertEqual(nfg_text, expected_rps_nfg)

  def test_pd(self):
    expected_pd_nfg = ("""NFG 1 R "OpenSpiel export of matrix_pd()"
{ "Player 0" "Player 1" } { 2 2 }

5 5
10 0
0 10
1 1
""")
    game = pyspiel.load_game("matrix_pd")
    nfg_text = pyspiel.game_to_nfg_string(game)
    self.assertEqual(nfg_text, expected_pd_nfg)

  def test_mp3p(self):
    expected_mp3p_nfg = ("""NFG 1 R "OpenSpiel export of matching_pennies_3p()"
{ "Player 0" "Player 1" "Player 2" } { 2 2 2 }

1 1 -1
-1 1 1
-1 -1 -1
1 -1 1
1 -1 1
-1 -1 -1
-1 1 1
1 1 -1
""")
    game = pyspiel.load_game("matching_pennies_3p")
    nfg_text = pyspiel.game_to_nfg_string(game)
    self.assertEqual(nfg_text, expected_mp3p_nfg)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.adidas_utils.helpers.symmetric.exploitability.

Computing the exploitability of a tsallis-entropy regularized game is more
involved, so we include a derivation here of an example test case using the
prisoner's dilemma (see pd np.array below). Note that the tsallis-entropy
setting assumes non-negative payoffs so we add 3 to the array. We assume p=1
for the tsallis entropy in this example.

pd    dist   grad     br       payoff(br)     payoff(dist)
[2 0] [.5] = [1] --> [1/3] --> 5/3        --> 3/2
[3 1] [.5]   [2]     [2/3]

s = sum(grad) = 3

tsallis-entr(br) = s / (p + 1) * (1 - br_1^2 - br_2^2)
                 = 3 / 2 * (1 - 1/9 - 4/9) = 2/3

tsallis-entr(dist) = s / (p + 1) * (1 - dist_1^2 - dist_2^2)
                   = 3 / 2 * (1 - 1/4 - 1/4) = 3/4

u(br) - u(dist) = 5/3 + 2/3 - 3/2 - 3/4 = 7 / 3 - 9 / 4
"""

from absl import logging  # pylint:disable=unused-import
from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability


test_seed = 12345

# prisoner's dilemma test case
# pylint:disable=bad-whitespace
pt_r = np.array([[-1, -3],
                 [0,  -2]])
# pylint:enable=bad-whitespace
pt_r -= pt_r.min()
pt_c = pt_r.T
pd = np.stack((pt_r, pt_c), axis=0)
pd_nash = np.array([0, 1])
pd_non_nash_1 = np.array([1, 0])
pd_non_nash_exp_1 = 1.
pd_non_nash_ate_exp_1 = pd_non_nash_exp_1
pd_non_nash_2 = np.array([0.5, 0.5])
pd_non_nash_exp_2 = 0.5
pd_non_nash_ate_exp_2 = 7. / 3. - 9. / 4.

qre_br = np.exp([1, 2]) / np.exp([1, 2]).sum()
entr_br = -np.sum(qre_br * np.log(qre_br))
entr_non_nash_2 = -np.sum(pd_non_nash_2 * np.log(pd_non_nash_2))
u_br_minus_non_nash = (qre_br - pd_non_nash_2).dot([1, 2])
pd_non_nash_qre_exp_2 = u_br_minus_non_nash + (entr_br - entr_non_nash_2)

# rock-paper-scissors test case
# pylint:disable=bad-whitespace
pt_r = np.array([[0, -1,  1],
                 [1,  0, -1],
                 [-1, 1,  0]])
# pylint:enable=bad-whitespace
pt_r -= pt_r.min()
pt_c = pt_r.T
rps = np.stack((pt_r, pt_c), axis=0)
rps_nash = np.ones(3) / 3.
rps_non_nash_1 = np.array([1, 0, 0])
rps_non_nash_exp_1 = 1.
rps_non_nash_2 = np.array([0, 1, 0])
rps_non_nash_exp_2 = 1.
rps_non_nash_3 = np.array([0, 0, 1])
rps_non_nash_exp_3 = 1.


class ExploitabilityTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('PD_nash', pd, pd_nash),
      ('RPS_nash', rps, rps_nash),
      )
  def test_unreg_exploitability_of_nash(self, payoff_tensor, nash):
    # assumes symmetric games
    exp = exploitability.unreg_exploitability(nash, payoff_tensor)
    self.assertEqual(exp, 0., 'nash should have zero exploitability')

  @parameterized.named_parameters(
      ('PD_non_nash_1', pd, pd_non_nash_1, pd_non_nash_exp_1),
      ('PD_non_nash_2', pd, pd_non_nash_2, pd_non_nash_exp_2),
      ('RPS_non_nash_1', rps, rps_non_nash_1, rps_non_nash_exp_1),
      ('RPS_non_nash_2', rps, rps_non_nash_2, rps_non_nash_exp_2),
      ('RPS_non_nash_3', rps, rps_non_nash_3, rps_non_nash_exp_3),
      )
  def test_unreg_exploitability_of_non_nash(self, payoff_tensor, dist, exp):
    # assumes symmetric games
    exp_pred = exploitability.unreg_exploitability(dist, payoff_tensor)
    self.assertEqual(exp_pred, exp, 'dist should have the given exploitability')

  @parameterized.named_parameters(
      ('PD_rand', pd, test_seed),
      ('RPS_rand', rps, test_seed),
      )
  def test_unreg_exploitability_of_rand(self, payoff_tensor, seed=None):
    trials = 100
    random = np.random.RandomState(seed)
    num_strategies = payoff_tensor.shape[-1]
    dists = random.rand(trials, num_strategies)
    dists /= np.sum(dists, axis=1, keepdims=True)
    exploitable = []
    for dist in dists:
      exp = exploitability.unreg_exploitability(dist, payoff_tensor)
      exploitable.append(exp > 0.)
    perc = 100 * np.mean(exploitable)
    logging.info('rand strat exploitable rate out of %d is %f', trials, perc)
    self.assertEqual(perc, 100., 'found rand strat that was nash')

  @parameterized.named_parameters(
      ('RPS_nash_p=0', rps, rps_nash, 0.),
      ('RPS_nash_p=0.1', rps, rps_nash, 0.1),
      ('RPS_nash_p=1', rps, rps_nash, 1.),
      )
  def test_ate_exploitability_of_nash(self, payoff_tensor, nash, p):
    # assumes symmetric games
    exp = exploitability.ate_exploitability(nash, payoff_tensor, p)
    self.assertGreaterEqual(0., exp,
                            'uniform nash should have zero exploitability')

  @parameterized.named_parameters(
      ('PD_non_nash_p=0', pd, 0., pd_non_nash_1, pd_non_nash_exp_1),
      ('PD_non_nash_p=1', pd, 1., pd_non_nash_2, pd_non_nash_ate_exp_2),
      )
  def test_ate_exploitability_of_non_nash(self, payoff_tensor, p, dist, exp):
    # assumes symmetric games
    exp_pred = exploitability.ate_exploitability(dist, payoff_tensor, p)
    self.assertAlmostEqual(exp_pred, exp,
                           msg='dist should have the given exploitability')

  @parameterized.named_parameters(
      ('RPS_rand_p=0', rps, 0., test_seed),
      ('RPS_rand_p=0.1', rps, 0.1, test_seed),
      ('RPS_rand_p=1', rps, 1., test_seed),
      )
  def test_ate_exploitability_of_rand(self, payoff_tensor, p, seed=None):
    trials = 100
    random = np.random.RandomState(seed)
    num_strategies = payoff_tensor.shape[-1]
    dists = random.rand(trials, num_strategies)
    dists /= np.sum(dists, axis=1, keepdims=True)
    exploitable = []
    for dist in dists:
      exp = exploitability.ate_exploitability(dist, payoff_tensor, p)
      exploitable.append(exp > 0.)
    perc = 100 * np.mean(exploitable)
    logging.info('rand strat exploitable rate out of %d is %f', trials, perc)
    self.assertEqual(perc, 100., 'found rand strat that was nash')

  @parameterized.named_parameters(
      ('RPS_nash_tau=0', rps, rps_nash, 0.),
      ('RPS_nash_tau=0.1', rps, rps_nash, 0.1),
      ('RPS_nash_tau=1', rps, rps_nash, 1.),
      )
  def test_qre_exploitability_of_nash(self, payoff_tensor, nash, temperature):
    # assumes symmetric games
    exp = exploitability.qre_exploitability(nash, payoff_tensor, temperature)
    self.assertGreaterEqual(1e-10, exp,
                            'uniform nash should have zero exploitability')

  @parameterized.named_parameters(
      ('PD_non_nash_tau=0', pd, 0., pd_non_nash_1, pd_non_nash_exp_1),
      ('PD_non_nash_tau=1', pd, 1., pd_non_nash_2, pd_non_nash_qre_exp_2),
      )
  def test_qre_exploitability_of_non_nash(self, payoff_tensor, temperature,
                                          dist, exp):
    # assumes symmetric games
    exp_pred = exploitability.qre_exploitability(dist, payoff_tensor,
                                                 temperature)
    self.assertAlmostEqual(exp_pred, exp,
                           msg='dist should have the given exploitability')

  @parameterized.named_parameters(
      ('RPS_rand_tau=0', rps, 0., test_seed),
      ('RPS_rand_tau=0.1', rps, 0.1, test_seed),
      ('RPS_rand_tau=1', rps, 1., test_seed),
      )
  def test_qre_exploitability_of_rand(self, payoff_tensor, temperature,
                                      seed=None):
    trials = 100
    random = np.random.RandomState(seed)
    num_strategies = payoff_tensor.shape[-1]
    dists = random.rand(trials, num_strategies)
    dists /= np.sum(dists, axis=1, keepdims=True)
    exploitable = []
    for dist in dists:
      exp = exploitability.qre_exploitability(dist, payoff_tensor, temperature)
      exploitable.append(exp > 0.)
    perc = 100 * np.mean(exploitable)
    logging.info('rand strat exploitable rate out of %d is %f', trials, perc)
    self.assertEqual(perc, 100., 'found rand strat that was nash')


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.egt.visualization."""

from absl import logging
from absl.testing import absltest

# pylint: disable=g-import-not-at-top
try:
  from matplotlib.figure import Figure
  from matplotlib.quiver import Quiver
  from matplotlib.streamplot import StreamplotSet
except ImportError as e:
  logging.info("If your tests failed with the error 'ImportError: No module "
               "named functools_lru_cache', this is a known bug in matplotlib "
               "and there is a workaround (run sudo apt install "
               "python-backports.functools-lru-cache. See: "
               "https://github.com/matplotlib/matplotlib/issues/9344.")
  raise e

import numpy as np

from open_spiel.python.egt import dynamics
from open_spiel.python.egt import utils
from open_spiel.python.egt import visualization
import pyspiel


def _build_dynamics2x2():
  """Build multi-population dynamics."""
  game = pyspiel.load_game("matrix_pd")
  payoff_tensor = utils.game_payoffs_array(game)
  return dynamics.MultiPopulationDynamics(payoff_tensor, dynamics.replicator)


def _build_dynamics3x3():
  """Build single-population dynamics."""
  game = pyspiel.load_game("matrix_rps")
  payoff_tensor = utils.game_payoffs_array(game)
  return dynamics.SinglePopulationDynamics(payoff_tensor, dynamics.replicator)


def _identity_dynamics(x):
  """Returns same input as output."""
  return x


class VisualizationTest(absltest.TestCase):

  def test_meshgrid(self):
    n = 10
    payoff_tensor = np.ones(shape=(2, 2, 2))
    identity = lambda x, f: x
    allzero = lambda x, f: np.zeros(x.shape)
    dyn = dynamics.MultiPopulationDynamics(payoff_tensor, (identity, allzero))
    x, y, u, v = visualization._eval_dynamics_2x2_grid(dyn, n)
    np.testing.assert_allclose(x, u)
    np.testing.assert_allclose(v, np.zeros(shape=(n, n)))

    dyn = dynamics.MultiPopulationDynamics(payoff_tensor, (allzero, identity))
    x, y, u, v = visualization._eval_dynamics_2x2_grid(dyn, n)
    np.testing.assert_allclose(u, np.zeros(shape=(n, n)))
    np.testing.assert_allclose(y, v)

  def test_quiver2x2(self):
    """Test 2x2 quiver plot."""
    dyn = _build_dynamics2x2()
    fig = Figure(figsize=(4, 4))
    ax = fig.add_subplot(111, projection="2x2")
    res = ax.quiver(dyn)
    self.assertIsInstance(res, Quiver)

  def test_streamplot2x2(self):
    """Test 2x2 quiver plot."""
    dyn = _build_dynamics2x2()
    fig = Figure(figsize=(4, 4))
    ax = fig.add_subplot(111, projection="2x2")
    res = ax.streamplot(dyn)
    self.assertIsInstance(res, StreamplotSet)

  def test_quiver3x3(self):
    """Test 3x3 quiver plot."""
    dyn = _build_dynamics3x3()
    fig = Figure(figsize=(4, 4))
    ax = fig.add_subplot(111, projection="3x3")
    res = ax.quiver(dyn)
    self.assertIsInstance(res, Quiver)

  def test_streamplot3x3(self):
    """Test 3x3 quiver plot."""
    dyn = _build_dynamics3x3()
    fig = Figure(figsize=(4, 4))
    ax = fig.add_subplot(111, projection="3x3")
    res = ax.streamplot(dyn)
    self.assertIsInstance(res, visualization.SimplexStreamMask)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.noisy_policy."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import games  # pylint:disable=unused-import
from open_spiel.python import policy as openspiel_policy
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.algorithms import noisy_policy
import pyspiel


class NoisyPolicyTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(["kuhn_poker", "leduc_poker"])
  def test_cpp_and_python_implementations_are_identical(self, game_name):
    game = pyspiel.load_game(game_name)

    policy = openspiel_policy.UniformRandomPolicy(game)

    all_states = get_all_states.get_all_states(
        game,
        depth_limit=-1,
        include_terminals=False,
        include_chance_states=False,
        to_string=lambda s: s.information_state_string())

    for current_player in range(game.num_players()):
      noise = noisy_policy.NoisyPolicy(
          policy, player_id=current_player, alpha=0.5, beta=10.)
      for state in all_states.values():
        if state.current_player() < 0:
          continue

        if state.current_player() != current_player:
          self.assertEqual(
              policy.action_probabilities(state),
              noise.action_probabilities(state))
        else:
          self.assertNotEqual(
              policy.action_probabilities(state),
              noise.action_probabilities(state))

  @parameterized.parameters(["python_iterated_prisoners_dilemma"])
  def test_simultaneous_game_noisy_policy(self, game_name):
    game = pyspiel.load_game(game_name)

    policy = openspiel_policy.UniformRandomPolicy(game)

    all_states = get_all_states.get_all_states(
        game,
        depth_limit=10,
        include_terminals=False,
        include_chance_states=False,
        to_string=lambda s: s.history_str())

    for current_player in range(game.num_players()):
      noise = noisy_policy.NoisyPolicy(
          policy, player_id=current_player, alpha=0.5, beta=10.)
      for state in all_states.values():
        if state.current_player() == pyspiel.PlayerId.SIMULTANEOUS:
          for player_id in range(game.num_players()):
            if player_id != current_player:
              self.assertEqual(
                  policy.action_probabilities(state, player_id),
                  noise.action_probabilities(state, player_id))
            else:
              self.assertNotEqual(
                  policy.action_probabilities(state, player_id),
                  noise.action_probabilities(state, player_id))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Neural Replicator Dynamics [Omidshafiei et al, 2019].

A policy gradient-like extension to replicator dynamics and the hedge algorithm
that incorporates function approximation.

# References

Shayegan Omidshafiei, Daniel Hennes, Dustin Morrill, Remi Munos,
  Julien Perolat, Marc Lanctot, Audrunas Gruslys, Jean-Baptiste Lespiau,
  Karl Tuyls. Neural Replicator Dynamics. https://arxiv.org/abs/1906.00190.
  2019.
"""

import numpy as np
import torch
from torch import nn

from open_spiel.python.pytorch import rcfr


def thresholded(logits, regrets, threshold=2.0):
  """Zeros out `regrets` where `logits` are too negative or too large."""
  can_decrease = torch.gt(logits, -threshold).float()
  can_increase = torch.lt(logits, threshold).float()
  regrets_negative = torch.minimum(regrets, torch.Tensor([0.0]))
  regrets_positive = torch.maximum(regrets, torch.Tensor([0.0]))
  return can_decrease * regrets_negative + can_increase * regrets_positive


def train(model,
          data,
          batch_size,
          step_size=1.0,
          threshold=2.0,
          autoencoder_loss=None):
  """Train NeuRD `model` on `data`."""
  data = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=True)

  for x, regrets in data:
    output = model(x, training=True)
    logits = output[:, :1]
    logits = logits - torch.mean(logits)

    regrets = thresholded(logits, regrets, threshold=threshold).detach()
    utility = torch.mean(logits * regrets)

    if autoencoder_loss is not None:
      utility = utility - autoencoder_loss(x, output[:, 1:])
    model.zero_grad()
    utility.backward()
    with torch.no_grad():
      for var in model.layers.parameters():
        new_var = var + step_size * var.grad
        var.copy_(new_var)


class DeepNeurdModel(nn.Module):
  """A flexible deep feedforward NeuRD model class.

  Properties:
    layers: The `torch.nn.Linear` layers describing this model.
  """

  def __init__(self,
               game,
               num_hidden_units,
               num_hidden_layers=1,
               num_hidden_factors=0,
               hidden_activation=nn.ReLU,
               use_skip_connections=False,
               autoencode=False):
    """Creates a new `DeepNeurdModel.

    Args:
      game: The OpenSpiel game being solved.
      num_hidden_units: The number of units in each hidden layer.
      num_hidden_layers: The number of hidden layers. Defaults to 1.
      num_hidden_factors: The number of hidden factors or the matrix rank of the
        layer. If greater than zero, hidden layers will be split into two
        separate linear transformations, the first with
        `num_hidden_factors`-columns and the second with
        `num_hidden_units`-columns. The result is that the logical hidden layer
        is a rank-`num_hidden_units` matrix instead of a rank-`num_hidden_units`
        matrix. When `num_hidden_units < num_hidden_units`, this is effectively
        implements weight sharing. Defaults to 0.
      hidden_activation: The activation function to apply over hidden layers.
        Defaults to `torch.nn.Relu`.
      use_skip_connections: Whether or not to apply skip connections (layer
        output = layer(x) + x) on hidden layers. Zero padding or truncation is
        used to match the number of columns on layer inputs and outputs.
      autoencode: Whether or not to output a reconstruction of the inputs upon
        being called. Defaults to `False`.
    """
    super(DeepNeurdModel, self).__init__()
    self._autoencode = autoencode
    self._use_skip_connections = use_skip_connections
    self._hidden_are_factored = num_hidden_factors > 0

    self.layers = nn.ModuleList()
    self.input_size = rcfr.num_features(game)
    for _ in range(num_hidden_layers):
      if self._hidden_are_factored:
        self.layers.append(
            nn.Linear(self.input_size, num_hidden_factors, bias=True))
        self.input_size = num_hidden_factors

      self.layers.append(
          nn.Linear(self.input_size, num_hidden_units, bias=True))
      if hidden_activation:
        self.layers.append(hidden_activation())
      self.input_size = num_hidden_units

    self.layers.append(
        nn.Linear(
            self.input_size,
            1 + self._autoencode * rcfr.num_features(game),
            bias=True))

  def forward(self, x, training=False):
    """Evaluates this model on x.

    Args:
      x: Model input.
      training: Whether or not this is being called during training. If
        `training` and the constructor argument `autoencode` was `True`, then
        the output will contain the estimated regrets concatenated with a
        reconstruction of the input, otherwise only regrets will be returned.
        Defaults to `False`.

    Returns:
      The `torch.Tensor` resulting from evaluating this model on `x`. If
        `training` and the constructor argument `autoencode` was `True`, then
        it will contain the estimated regrets concatenated with a
        reconstruction of the input, otherwise only regrets will be returned.
    """
    y = rcfr.feedforward_evaluate(
        layers=self.layers,
        x=x,
        use_skip_connections=self._use_skip_connections,
        hidden_are_factored=self._hidden_are_factored)
    return y if training else y[:, :1]


class CounterfactualNeurdSolver(object):
  """All-actions, strong NeuRD on counterfactual regrets.

  No regularization bonus is applied, so the current policy likely will not
  converge. The average policy profile is updated and stored in a full
  game-size table and may converge to an approximate Nash equilibrium in
  two-player, zero-sum games.
  """

  def __init__(self, game, models):
    """Creates a new `CounterfactualNeurdSolver`.

    Args:
      game: An OpenSpiel `Game`.
      models: Current policy models (optimizable array-like -> `torch.Tensor`
        callables) for both players.
    """
    self._game = game
    self._models = models
    self._root_wrapper = rcfr.RootStateWrapper(game.new_initial_state())

    self._cumulative_seq_probs = [
        np.zeros(n) for n in self._root_wrapper.num_player_sequences
    ]

  def _sequence_weights(self, player=None):
    """Returns exponentiated weights for each sequence as an `np.array`."""
    if player is None:
      return [
          self._sequence_weights(player)
          for player in range(self._game.num_players())
      ]
    else:
      tensor = torch.squeeze(self._models[player](
          self._root_wrapper.sequence_features[player]))
      tensor = tensor - torch.max(tensor, dim=0)[0]
      tensor = torch.exp(tensor)
      return tensor.detach().numpy()

  def current_policy(self):
    """Returns the current policy profile.

    Returns:
      A `dict<info state, list<Action, probability>>` that maps info state
      strings to `Action`-probability pairs describing each player's policy.
    """
    return self._root_wrapper.sequence_weights_to_tabular_profile(
        self._sequence_weights())

  def average_policy(self):
    """Returns the average of all policies iterated.

    The policy is computed using the accumulated policy probabilities computed
    using `evaluate_and_update_policy`.

    Returns:
      A `dict<info state, list<Action, probability>>` that maps info state
      strings to (Action, probability) pairs describing each player's policy.
    """
    return self._root_wrapper.sequence_weights_to_tabular_profile(
        self._cumulative_seq_probs)

  def _previous_player(self, player):
    """The previous player in the turn ordering."""
    return player - 1 if player > 0 else self._game.num_players() - 1

  def _average_policy_update_player(self, regret_player):
    """The player for whom the average policy should be updated."""
    return self._previous_player(regret_player)

  def evaluate_and_update_policy(self, train_fn):
    """Performs a single step of policy evaluation and policy improvement.

    Args:
      train_fn: A (model, `torch.utils.data.TensorDataset`) function that trains
        the given regression model to accurately reproduce the x to y mapping
        given x-y data.
    """
    sequence_weights = self._sequence_weights()
    player_seq_features = self._root_wrapper.sequence_features
    for regret_player in range(self._game.num_players()):
      seq_prob_player = self._average_policy_update_player(regret_player)

      regrets, seq_probs = (
          self._root_wrapper.counterfactual_regrets_and_reach_weights(
              regret_player, seq_prob_player, *sequence_weights))

      self._cumulative_seq_probs[seq_prob_player] += seq_probs
      targets = torch.unsqueeze(torch.Tensor(regrets), axis=1)
      data = torch.utils.data.TensorDataset(player_seq_features[regret_player],
                                            targets)

      regret_player_model = self._models[regret_player]
      train_fn(regret_player_model, data)
      sequence_weights[regret_player] = self._sequence_weights(regret_player)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.adidas_utils.solvers.symmetric."""

from absl import logging  # pylint:disable=unused-import
from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from scipy.spatial.distance import cosine

from open_spiel.python.algorithms.adidas_utils.solvers.symmetric import ate
from open_spiel.python.algorithms.adidas_utils.solvers.symmetric import ped
from open_spiel.python.algorithms.adidas_utils.solvers.symmetric import qre


class ExploitabilityDescentTest(parameterized.TestCase):

  @staticmethod
  def numerical_gradient(fun, x, eps=np.sqrt(np.finfo(float).eps)):
    fun_0 = fun(x)
    num_grad = np.zeros_like(x)
    x_plus_dx = np.copy(x)
    for i in range(len(x)):
      x_plus_dx[i] = x[i] + eps
      num_grad[i] = (fun(x_plus_dx) - fun_0) / eps
      x_plus_dx[i] = x[i]
    return num_grad

  @staticmethod
  def prep_params(dist, payoff_matrices, num_params, solver_tuple):
    params = [dist]
    if num_params > 1:
      params += [payoff_matrices[0].dot(params[0])]  # policy_gradient
    if num_params > 2:
      params += [np.linalg.norm(params[1], ord=solver_tuple[1])]
    return tuple(params)

  @parameterized.named_parameters(
      ("PED", (ped, False)),
      ("ATE_p=1", (ate, 1., False)),
      ("ATE_p=0.5", (ate, 0.5, False)),
      ("ATE_p=0.1", (ate, 0.1, False)),
      ("QRE_t=0.0", (qre, 0.0, False)),
      ("QRE_t=0.1", (qre, 0.1, False))
      )
  def test_exploitability_gradient_on_symmetric_two_player_matrix_games(
      self, solver_tuple, trials=100, max_num_strats=2, atol=1e-1, rtol=1e-1,
      seed=1234):
    num_players = 2
    solver = solver_tuple[0].Solver(*solver_tuple[1:])

    random = np.random.RandomState(seed)

    successes = []
    for _ in range(trials):
      num_strats = random.randint(low=2, high=max_num_strats + 1)
      strat_dims = (num_strats,) * num_players
      payoff_matrices = random.rand(num_players, *strat_dims)
      payoff_matrices[1] = payoff_matrices[0].T

      num_params = len(solver.init_vars(num_strats, num_players))
      dirichlet_alpha = np.ones(num_strats)
      dist = random.dirichlet(dirichlet_alpha)  # mixed srategy
      params = self.prep_params(dist, payoff_matrices, num_params, solver_tuple)

      grad = solver.compute_gradients(params, payoff_matrices)[0][0]

      exp = lambda x: solver.exploitability(x, payoff_matrices)  # pylint: disable=cell-var-from-loop
      num_grad = self.numerical_gradient(exp, dist)

      successes += [np.logical_and(np.allclose(grad, num_grad, rtol, atol),
                                   cosine(grad, num_grad) <= atol)]

    perc = 100 * np.mean(successes)
    logging.info("gradient accuracy success rate out of %d is %f", trials, perc)
    self.assertGreaterEqual(
        perc, 95., "exploitability gradient accuracy is too poor")


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.cfr."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import outcome_sampling_mccfr
import pyspiel

# Convergence results change depending on
# the seed specified for running the tests.
# For this reason, test thresholds have been adapted
# taking the maximum Nash exploitability value obtained
# from multiple runs.
# For more details see https://github.com/deepmind/open_spiel/pull/458
SEED = 39823987


class OutcomeSamplingMCCFRTest(absltest.TestCase):

  def test_outcome_sampling_leduc_2p(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("leduc_poker")
    os_solver = outcome_sampling_mccfr.OutcomeSamplingSolver(game)
    for _ in range(10000):
      os_solver.iteration()
    conv = exploitability.nash_conv(game, os_solver.average_policy())
    print("Leduc2P, conv = {}".format(conv))

    self.assertLess(conv, 3.07)

  def test_outcome_sampling_kuhn_2p(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker")
    os_solver = outcome_sampling_mccfr.OutcomeSamplingSolver(game)
    for _ in range(10000):
      os_solver.iteration()
    conv = exploitability.nash_conv(game, os_solver.average_policy())
    print("Kuhn2P, conv = {}".format(conv))
    self.assertLess(conv, 0.17)
    # ensure that to_tabular() works on the returned policy
    # and the tabular policy is equivalent
    tabular_policy = os_solver.average_policy().to_tabular()
    conv2 = exploitability.nash_conv(game, tabular_policy)
    self.assertEqual(conv, conv2)

  def test_outcome_sampling_kuhn_3p(self):
    np.random.seed(SEED)
    game = pyspiel.load_game("kuhn_poker", {"players": 3})
    os_solver = outcome_sampling_mccfr.OutcomeSamplingSolver(game)
    for _ in range(10000):
      os_solver.iteration()
    conv = exploitability.nash_conv(game, os_solver.average_policy())
    print("Kuhn3P, conv = {}".format(conv))
    self.assertLess(conv, 0.22)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Solving strong Stackelberg equilibrium based on linear programming.

Based on [1] "Computing the Optimal Strategy to Commit to", Conitzer & Sandholm,
EC'06
"""

import numpy as np

from open_spiel.python.algorithms import lp_solver
from open_spiel.python.algorithms.projected_replicator_dynamics import _simplex_projection
from open_spiel.python.egt.utils import game_payoffs_array


def solve_stackelberg(game, is_first_leader=True):
  """Solves the optimal mixed strategty to commit to for the leader.

  Args:
    game: a pyspiel game,
    is_first_leader: if true, then player 0 is the leader, o.w. player 1 is
      the leader.

  Returns:
    (player0 strategy, player1 strategy, player0 payoff, player1 payoff) at an
    SSE.
  """
  p_mat = game_payoffs_array(game)
  assert len(p_mat) == 2
  if is_first_leader:
    leader_payoff, follower_payoff = p_mat[0], p_mat[1]
  else:
    leader_payoff, follower_payoff = p_mat[1].T, p_mat[0].T

  num_leader_strategies, num_follower_strategies = leader_payoff.shape

  leader_eq_value = -float("inf")
  follower_eq_value = None
  leader_eq_strategy = None
  follower_eq_strategy = None

  for t in range(num_follower_strategies):
    lp = lp_solver.LinearProgram(objective=lp_solver.OBJ_MAX)
    for s in range(num_leader_strategies):
      lp.add_or_reuse_variable("s_{}".format(s))
      lp.set_obj_coeff("s_{}".format(s), leader_payoff[s, t])

    for t_ in range(num_follower_strategies):
      if t_ == t:
        continue
      lp.add_or_reuse_constraint("t_{}".format(t_), lp_solver.CONS_TYPE_GEQ)
      for s in range(num_leader_strategies):
        lp.set_cons_coeff("t_{}".format(t_), "s_{}".format(s),
                          follower_payoff[s, t] - follower_payoff[s, t_])
      lp.set_cons_rhs("t_{}".format(t_), 0.0)
    lp.add_or_reuse_constraint("sum_to_one", lp_solver.CONS_TYPE_EQ)
    for s in range(num_leader_strategies):
      lp.set_cons_coeff("sum_to_one", "s_{}".format(s), 1.0)
    lp.set_cons_rhs("sum_to_one", 1.0)
    try:
      leader_strategy = np.array(lp.solve())
      leader_strategy = _simplex_projection(
          leader_strategy.reshape(-1)).reshape(-1, 1)
      leader_value = leader_strategy.T.dot(leader_payoff)[0, t]
      if leader_value > leader_eq_value:
        leader_eq_strategy = leader_strategy
        follower_eq_strategy = t
        leader_eq_value = leader_value
        follower_eq_value = leader_strategy.T.dot(follower_payoff)[0, t]
    except:  # pylint: disable=bare-except
      continue
  if is_first_leader:
    return (leader_eq_strategy.reshape(-1), np.identity(
        num_follower_strategies)[follower_eq_strategy],
            leader_eq_value, follower_eq_value)
  else:
    return (np.identity(num_follower_strategies)[follower_eq_strategy],
            leader_eq_strategy.reshape(-1), follower_eq_value, leader_eq_value)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MFG utilities."""

import pickle

from open_spiel.python.utils import gfile
from open_spiel.python.mfg import distribution


def save_parametric_distribution(dist: distribution.ParametricDistribution,
                                 filename: str):
  """Saves the parametric distribution to a Pickle file."""
  with gfile.Open(filename, "wb") as f:
    pickle.dump(dist.get_params(), f, protocol=pickle.DEFAULT_PROTOCOL)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""

import logging
import sys
from absl import app
from absl import flags
import numpy as np

import pyspiel
import tensorflow.compat.v1 as tf
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import dqn
from open_spiel.python.algorithms import random_agent
from open_spiel.python.algorithms import tabular_qlearner

from open_spiel.python import games  # pylint: disable=unused-import
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", int(1e1), "Number of train episodes.")
flags.DEFINE_boolean(
    "interactive_play",
    True,
    "Whether to run an interactive play with the agent after training.",
)

"""
_DICT_CHESS_CELL = {6: "雷", 5: "师", 4: "旅", 3: "团", 2: "炸", 1: "棋", 0: "空"}
def pretty_board(time_step):
    info_state = time_step.observations["info_state"]
    info_board = np.reshape(info_state[0], (13,6,3))
    board = [[f"\033[;;m空\033[0m"]*3]*6
    for chess_axis in range(1, len(info_board)):
        for i in range(6):
            for j in range(3):
                if info_board[chess_axis][i][j]:
                    if chess_axis >= 6:
                        axis = chess_axis - 6
                        board[i][j] = f"\033[;30;42m{_DICT_CHESS_CELL[axis]}\033[0m"
                    else:
                        axis = chess_axis + 0
                        board[i][j] = f"\033[;30;43m{_DICT_CHESS_CELL[axis]}\033[0m"
    string_board = "\n".join("".join([str(chess) for chess in row]) for row in board)
    return string_board
"""


def command_line_action(time_step):
    """Gets a valid action from the user on the command line."""
    current_player = time_step.observations["current_player"]
    legal_actions = time_step.observations["legal_actions"][current_player]
    action = -1
    while action not in legal_actions:
        print("Choose an action from {}:".format(legal_actions))
        sys.stdout.flush()
        action_str = input()
        try:
            action = int(action_str)
        except ValueError:
            continue
    return action


def eval_against_random_bots(env, trained_agents, random_agents, num_episodes):
    """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
    wins = np.zeros(2)
    for player_pos in range(2):
        if player_pos == 0:
            cur_agents = [trained_agents[0], random_agents[1]]
        else:
            cur_agents = [random_agents[0], trained_agents[1]]
        for _ in range(num_episodes):
            time_step = env.reset()
            while not time_step.last():
                player_id = time_step.observations["current_player"]
                agent_output = cur_agents[player_id].step(time_step, is_evaluation=True)
                time_step = env.step([agent_output.action])
            if time_step.rewards[player_pos] > 0:
                wins[player_pos] += 1
    return wins / num_episodes


class JunQiEnv(rl_environment.Environment):
    def get_stringed_board(self):
        return self._state.serialize()

    def get_stringed_action(self):
        actions = self._state.legal_actions()
        return "\n".join(self._state.serialize_action(action) for action in actions)


def main(_):
    game = "junqi1"
    num_players = 2
    env = JunQiEnv(game)
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    hidden_layers_sizes = [32, 32]
    replay_buffer_capacity = int(1e4)
    train_episodes = FLAGS.num_episodes
    loss_report_interval = 1000

    with tf.Session() as sess:
        dqn_agent = dqn.DQN(
            sess,
            player_id=0,
            state_representation_size=state_size,
            num_actions=num_actions,
            hidden_layers_sizes=hidden_layers_sizes,
            replay_buffer_capacity=replay_buffer_capacity)
        tabular_q_agent = tabular_qlearner.QLearner(
            player_id=1, num_actions=num_actions)
        agents = [dqn_agent, tabular_q_agent]

        sess.run(tf.global_variables_initializer())

        # Train agent

        for ep in range(train_episodes):
            if ep and ep % loss_report_interval == 0:
                logging.info("[%s/%s] DQN loss: %s", ep, train_episodes, agents[0].loss)
            time_step = env.reset()
            while not time_step.last():
                player_id = time_step.observations["current_player"]
                agent_output = agents[player_id].step(time_step)
                action_list = [agent_output.action]
                time_step = env.step(action_list)

            # Episode is over, step all agents with final info state.
            for agent in agents:
                agent.step(time_step)

        # Evaluate against random agent
        random_agents = [
            random_agent.RandomAgent(player_id=idx, num_actions=num_actions)
            for idx in range(num_players)
        ]
        r_mean = eval_against_random_bots(env, agents, random_agents, 1000)
        logging.info("Mean episode rewards: %s", r_mean)

        if not FLAGS.interactive_play:
            return

        # 2. Play from the command line against the trained agent.
        human_player = 0
        while True:
            print("You are playing as %s", "O" if human_player else "X")
            time_step = env.reset()
            while not time_step.last():
                player_id = time_step.observations["current_player"]
                if player_id == human_player:
                    agent_out = agents[human_player].step(time_step, is_evaluation=True)
                    # print("\n%s", agent_out)
                    print(env.get_stringed_board(), end="\n\n")
                    print(env.get_stringed_action())
                    action = command_line_action(time_step)
                else:
                    agent_out = agents[1 - human_player].step(time_step, is_evaluation=True)
                    action = agent_out.action
                time_step = env.step([action])

            print(env.get_stringed_board(), end="\n\n")

            print("End of game!")
            if time_step.rewards[human_player] > 0:
                print("You win")
            elif time_step.rewards[human_player] < 0:
                print("You lose")
            else:
                print("Draw")
            # Switch order of players
            human_player = 1 - human_player


if __name__ == "__main__":
    app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some helpers for normal-form games."""

import collections
import numpy as np


class StrategyAverager(object):
  """A helper class for averaging strategies for players."""

  def __init__(self, num_players, action_space_shapes, window_size=None):
    """Initialize the average strategy helper object.

    Args:
      num_players (int): the number of players in the game,
      action_space_shapes:  an vector of n integers, where each element
          represents the size of player i's actions space,
      window_size (int or None): if None, computes the players' average
          strategies over the entire sequence, otherwise computes the average
          strategy over a finite-sized window of the k last entries.
    """
    self._num_players = num_players
    self._action_space_shapes = action_space_shapes
    self._window_size = window_size
    self._num = 0
    if self._window_size is None:
      self._sum_meta_strategies = [
          np.zeros(action_space_shapes[p]) for p in range(num_players)
      ]
    else:
      self._window = collections.deque(maxlen=self._window_size)

  def append(self, meta_strategies):
    """Append the meta-strategies to the averaged sequence.

    Args:
      meta_strategies: a list of strategies, one per player.
    """
    if self._window_size is None:
      for p in range(self._num_players):
        self._sum_meta_strategies[p] += meta_strategies[p]
    else:
      self._window.append(meta_strategies)
    self._num += 1

  def average_strategies(self):
    """Return each player's average strategy.

    Returns:
      The averaged strategies, as a list containing one strategy per player.
    """

    if self._window_size is None:
      avg_meta_strategies = [
          np.copy(x) for x in self._sum_meta_strategies
      ]
      num_strategies = self._num
    else:
      avg_meta_strategies = [
          np.zeros(self._action_space_shapes[p])
          for p in range(self._num_players)
      ]
      for i in range(len(self._window)):
        for p in range(self._num_players):
          avg_meta_strategies[p] += self._window[i][p]
      num_strategies = len(self._window)
    for p in range(self._num_players):
      avg_meta_strategies[p] /= num_strategies
    return avg_meta_strategies



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for meta learning for regret minimization."""

from absl import flags
import jax
import jax.numpy as jnp
import numpy as np

FLAGS = flags.FLAGS


def meta_loss(opt_params, net_apply, payoff, steps):

  """Returns the meta learning loss value.

  Args:
    opt_params: Optimizer parameters.
    net_apply: Apply function.
    payoff: Payoff matrix.
    steps: Number of steps.

  Returns:
    Accumulated loss value over number of steps.

  """
  regret_sum_x = np.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])
  regret_sum_y = np.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])
  total_loss = 0
  step = 0

  @jax.jit
  def scan_body(carry, x):
    nonlocal regret_sum_x
    nonlocal regret_sum_y
    regret_sum_x, regret_sum_y, current_step, total_loss = carry
    x = net_apply(opt_params, None, regret_sum_x / (current_step + 1))
    y = net_apply(opt_params, None, regret_sum_y / (current_step + 1))

    strategy_x = jax.nn.softmax(x)
    strategy_y = jnp.transpose(jax.nn.softmax(y), [0, 2, 1])

    values_x = jnp.matmul(payoff, strategy_y)  # val_x = payoff * st_y
    values_y = -jnp.matmul(strategy_x, payoff)  # val_y = -1 * payoff * st_x

    value_x = jnp.matmul(jnp.matmul(strategy_x, payoff), strategy_y)
    value_y = -value_x

    curren_regret_x = values_x - value_x
    curren_regret_y = values_y - value_y
    curren_regret_x = jnp.transpose(curren_regret_x, [0, 2, 1])

    regret_sum_x += curren_regret_x
    regret_sum_y += curren_regret_y

    current_loss = jnp.mean(jnp.max(
        jax.numpy.concatenate([curren_regret_x, curren_regret_y], axis=2),
        axis=[1, 2]), axis=-1)
    total_loss += current_loss
    current_step += 1
    return (regret_sum_x, regret_sum_y, current_step, total_loss), None

  (regret_sum_x, regret_sum_y, step, total_loss), _ = jax.lax.scan(
      scan_body,
      (regret_sum_x, regret_sum_y, step, total_loss),
      None,
      length=steps,
  )

  return total_loss



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.environment.cliff_walking."""

import random
from absl.testing import absltest
from open_spiel.python import rl_environment
from open_spiel.python.environments import cliff_walking


def _select_random_legal_action(time_step):
  cur_legal_actions = time_step.observations["legal_actions"][0]
  action = random.choice(cur_legal_actions)
  return action


class CliffWalkingEnvTest(absltest.TestCase):

  def test_obs_spec(self):
    env = cliff_walking.Environment()
    obs_specs = env.observation_spec()
    self.assertLen(obs_specs, 3)
    self.assertCountEqual(obs_specs.keys(),
                          ["current_player", "info_state", "legal_actions"])
    self.assertEqual(obs_specs["info_state"], (2,))

  def test_action_spec(self):
    env = cliff_walking.Environment()
    action_spec = env.action_spec()
    self.assertLen(action_spec, 4)
    self.assertCountEqual(action_spec.keys(),
                          ["dtype", "max", "min", "num_actions"])
    self.assertEqual(action_spec["num_actions"], 4)
    self.assertEqual(action_spec["dtype"], int)

  def test_action_interfaces(self):
    env = cliff_walking.Environment()
    time_step = env.reset()

    # Singleton list works
    action_list = [cliff_walking.UP]
    time_step = env.step(action_list)
    self.assertEqual(time_step.step_type, rl_environment.StepType.MID)

    # Integer works
    action_int = cliff_walking.UP
    time_step = env.step(action_int)
    self.assertEqual(time_step.step_type, rl_environment.StepType.MID)

  def test_many_runs(self):
    random.seed(1234)
    for _ in range(30):
      height = random.randint(3, 10)
      width = random.randint(3, 10)
      env = cliff_walking.Environment(height=height, width=width)

      time_step = env.reset()
      self.assertEqual(time_step.step_type, rl_environment.StepType.FIRST)
      self.assertIsNone(time_step.rewards)

      action_int = cliff_walking.UP
      time_step = env.step(action_int)
      self.assertEqual(time_step.step_type, rl_environment.StepType.MID)
      self.assertEqual(time_step.rewards, [-1.0])

      action_int = cliff_walking.RIGHT
      for _ in range(1, width):
        time_step = env.step(action_int)
        self.assertEqual(time_step.step_type, rl_environment.StepType.MID)
        self.assertEqual(time_step.rewards, [-1.0])

      action_int = cliff_walking.DOWN
      time_step = env.step(action_int)

      self.assertEqual(time_step.step_type, rl_environment.StepType.LAST)
      self.assertEqual(time_step.rewards, [-1.0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Discounted CFR and Linear CFR algorithms.

This implements Discounted CFR and Linear CFR, from Noam Brown and Tuomas
Sandholm, 2019, "Solving Imperfect-Information Games via Discounted Regret
Minimization".
See https://arxiv.org/abs/1809.04040.

Linear CFR (LCFR), is identical to CFR, except on iteration `t` the updates to
the regrets and average strategies are given weight `t`. (Equivalently, one
could multiply the accumulated regret by t / (t+1) on each iteration.)

Discounted CFR(alpha, beta, gamma) is defined by, at iteration `t`:
- multiplying the positive accumulated regrets by (t^alpha / (t^alpha + 1))
- multiplying the negative accumulated regrets by (t^beta / (t^beta + 1))
- multiplying the contribution to the average strategy by t^gamma

WARNING: This was contributed on Github, and the OpenSpiel team is not aware it
has been verified we can reproduce the paper results.
"""

import numpy as np

from open_spiel.python.algorithms import cfr

_InfoStateNode = cfr._InfoStateNode  # pylint: disable=protected-access


class _DCFRSolver(cfr._CFRSolver):  # pylint: disable=protected-access
  """Discounted CFR."""

  def __init__(self, game, alternating_updates, linear_averaging,
               regret_matching_plus, alpha, beta, gamma):
    super(_DCFRSolver, self).__init__(game, alternating_updates,
                                      linear_averaging, regret_matching_plus)
    self.alpha = alpha
    self.beta = beta
    self.gamma = gamma

    # We build a list of the nodes for all players, which will be updated
    # within `evaluate_and_update_policy`.
    self._player_nodes = [[] for _ in range(self._num_players)]
    for info_state in self._info_state_nodes.values():
      self._player_nodes[info_state.player].append(info_state)

  def _initialize_info_state_nodes(self, state):
    """Initializes info_state_nodes.

    We override the parent function, to add the current player information
    at the given node. This is used because we want to do updates for all nodes
    for a specific player.

    Args:
      state: The current state in the tree walk. This should be the root node
        when we call this function from a CFR solver.
    """
    if state.is_terminal():
      return

    if state.is_chance_node():
      for action, unused_action_prob in state.chance_outcomes():
        self._initialize_info_state_nodes(state.child(action))
      return

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)

    info_state_node = self._info_state_nodes.get(info_state)
    if info_state_node is None:
      legal_actions = state.legal_actions(current_player)
      info_state_node = _InfoStateNode(
          legal_actions=legal_actions,
          index_in_tabular_policy=self._current_policy.state_lookup[info_state])
      info_state_node.player = current_player
      self._info_state_nodes[info_state] = info_state_node

    for action in info_state_node.legal_actions:
      self._initialize_info_state_nodes(state.child(action))

  def _compute_counterfactual_regret_for_player(self, state, policies,
                                                reach_probabilities, player):
    """Increments the cumulative regrets and policy for `player`.

    Args:
      state: The initial game state to analyze from.
      policies: Unused. To be compatible with the `_CFRSolver` signature.
      reach_probabilities: The probability for each player of reaching `state`
        as a numpy array [prob for player 0, for player 1,..., for chance].
        `player_reach_probabilities[player]` will work in all cases.
      player: The 0-indexed player to update the values for. If `None`, the
        update for all players will be performed.

    Returns:
      The utility of `state` for all players, assuming all players follow the
      current policy defined by `self.Policy`.
    """
    if state.is_terminal():
      return np.asarray(state.returns())

    if state.is_chance_node():
      state_value = 0.0
      for action, action_prob in state.chance_outcomes():
        assert action_prob > 0
        new_state = state.child(action)
        new_reach_probabilities = reach_probabilities.copy()
        new_reach_probabilities[-1] *= action_prob
        state_value += action_prob * self._compute_counterfactual_regret_for_player(
            new_state, policies, new_reach_probabilities, player)
      return state_value

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)

    # No need to continue on this history branch as no update will be performed
    # for any player.
    # The value we return here is not used in practice. If the conditional
    # statement is True, then the last taken action has probability 0 of
    # occurring, so the returned value is not impacting the parent node value.
    if all(reach_probabilities[:-1] == 0):
      return np.zeros(self._num_players)

    state_value = np.zeros(self._num_players)

    # The utilities of the children states are computed recursively. As the
    # regrets are added to the information state regrets for each state in that
    # information state, the recursive call can only be made once per child
    # state. Therefore, the utilities are cached.
    children_utilities = {}

    info_state_node = self._info_state_nodes[info_state]
    if policies is None:
      info_state_policy = self._get_infostate_policy(info_state)
    else:
      info_state_policy = policies[current_player](info_state)
    for action in state.legal_actions():
      action_prob = info_state_policy.get(action, 0.)
      new_state = state.child(action)
      new_reach_probabilities = reach_probabilities.copy()
      new_reach_probabilities[current_player] *= action_prob
      child_utility = self._compute_counterfactual_regret_for_player(
          new_state,
          policies=policies,
          reach_probabilities=new_reach_probabilities,
          player=player)

      state_value += action_prob * child_utility
      children_utilities[action] = child_utility

    # If we are performing alternating updates, and the current player is not
    # the current_player, we skip the cumulative values update.
    # If we are performing simultaneous updates, we do update the cumulative
    # values.
    simulatenous_updates = player is None
    if not simulatenous_updates and current_player != player:
      return state_value

    reach_prob = reach_probabilities[current_player]
    counterfactual_reach_prob = (
        np.prod(reach_probabilities[:current_player]) *
        np.prod(reach_probabilities[current_player + 1:]))
    state_value_for_player = state_value[current_player]

    for action, action_prob in info_state_policy.items():
      cfr_regret = counterfactual_reach_prob * (
          children_utilities[action][current_player] - state_value_for_player)

      info_state_node = self._info_state_nodes[info_state]
      info_state_node.cumulative_regret[action] += cfr_regret
      if self._linear_averaging:
        info_state_node.cumulative_policy[action] += (
            reach_prob * action_prob * (self._iteration**self.gamma))
      else:
        info_state_node.cumulative_policy[action] += reach_prob * action_prob

    return state_value

  def evaluate_and_update_policy(self):
    """Performs a single step of policy evaluation and policy improvement."""
    self._iteration += 1
    if self._alternating_updates:
      for current_player in range(self._game.num_players()):
        self._compute_counterfactual_regret_for_player(
            self._root_node,
            policies=None,
            reach_probabilities=np.ones(self._game.num_players() + 1),
            player=current_player)
        for info_state in self._player_nodes[current_player]:
          for action in info_state.cumulative_regret.keys():
            if info_state.cumulative_regret[action] >= 0:
              info_state.cumulative_regret[action] *= (
                  self._iteration**self.alpha /
                  (self._iteration**self.alpha + 1))
            else:
              info_state.cumulative_regret[action] *= (
                  self._iteration**self.beta / (self._iteration**self.beta + 1))
        cfr._update_current_policy(self._current_policy, self._info_state_nodes)  # pylint: disable=protected-access


class DCFRSolver(_DCFRSolver):

  def __init__(self, game, alpha=3 / 2, beta=0, gamma=2):
    super(DCFRSolver, self).__init__(
        game,
        regret_matching_plus=False,
        alternating_updates=True,
        linear_averaging=True,
        alpha=alpha,
        beta=beta,
        gamma=gamma)


class LCFRSolver(_DCFRSolver):

  def __init__(self, game):
    super(LCFRSolver, self).__init__(
        game,
        regret_matching_plus=False,
        alternating_updates=True,
        linear_averaging=True,
        alpha=1,
        beta=1,
        gamma=1)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mirror Descent (https://arxiv.org/pdf/2103.00623.pdf)."""

from typing import Dict, List, Optional

import numpy as np

from open_spiel.python import policy as policy_lib
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import distribution
import pyspiel


def softmax_projection(logits):
  max_l = max(logits)
  exp_l = [np.exp(l - max_l) for l in logits]
  norm_exp = sum(exp_l)
  return [l / norm_exp for l in exp_l]


class ProjectedPolicy(policy_lib.Policy):
  """Project values on the policy simplex."""

  def __init__(
      self,
      game: pyspiel.Game,
      player_ids: List[int],
      state_value: value.ValueFunction,
      coeff: float = 1.0,
  ):
    """Initializes the projected policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      state_value: The (cumulative) state value to project.
      coeff: Coefficient for the values of the states.
    """
    super(ProjectedPolicy, self).__init__(game, player_ids)
    self._state_value = state_value
    self._coeff = coeff

  def value(self, state: pyspiel.State, action: Optional[int] = None) -> float:
    if action is None:
      return self._state_value(
          state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))
    else:
      new_state = state.child(action)
      return state.rewards()[0] + self._state_value(
          new_state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID))

  def action_probabilities(self,
                           state: pyspiel.State,
                           player_id: Optional[int] = None) -> Dict[int, float]:
    del player_id
    action_logit = [(a, self._coeff * self.value(state, action=a))
                    for a in state.legal_actions()]
    action, logit = zip(*action_logit)
    return dict(zip(action, softmax_projection(logit)))


class MirrorDescent(object):
  """The mirror descent algorithm."""

  def __init__(self,
               game: pyspiel.Game,
               state_value: Optional[value.ValueFunction] = None,
               lr: float = 0.01,
               root_state: Optional[pyspiel.State] = None):
    """Initializes mirror descent.

    Args:
      game: The game,
      state_value: A state value function. Default to TabularValueFunction.
      lr: The learning rate of mirror descent,
      root_state: The state of the game at which to start. If `None`, the game
        root state is used.
    """
    self._game = game
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._policy = policy_lib.UniformRandomPolicy(game)
    self._distribution = distribution.DistributionPolicy(game, self._policy)
    self._md_step = 0
    self._lr = lr

    self._state_value = (
        state_value if state_value else value.TabularValueFunction(game))
    self._cumulative_state_value = value.TabularValueFunction(game)

  def get_state_value(self, state: pyspiel.State,
                      learning_rate: float) -> float:
    """Returns the value of the state."""
    if state.is_terminal():
      return state.rewards()[state.mean_field_population()]

    if state.current_player() == pyspiel.PlayerId.CHANCE:
      v = 0.0
      for action, prob in state.chance_outcomes():
        new_state = state.child(action)
        v += prob * self.eval_state(new_state, learning_rate)
      return v

    if state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
      dist_to_register = state.distribution_support()
      dist = [
          self._distribution.value_str(str_state, 0.0)
          for str_state in dist_to_register
      ]
      new_state = state.clone()
      new_state.update_distribution(dist)
      return (state.rewards()[state.mean_field_population()] +
              self.eval_state(new_state, learning_rate))

    assert int(state.current_player()) >= 0, "The player id should be >= 0"
    v = 0.0
    for action, prob in self._policy.action_probabilities(state).items():
      new_state = state.child(action)
      v += prob * self.eval_state(new_state, learning_rate)
    return state.rewards()[state.mean_field_population()] + v

  def eval_state(self, state: pyspiel.State, learning_rate: float) -> float:
    """Evaluate the value of a state and update the cumulative sum."""
    state_str = state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    # Return the already calculated value if present.
    if self._state_value.has(state_str):
      return self._state_value(state_str)
    # Otherwise, calculate the value of the state.
    v = self.get_state_value(state, learning_rate)
    self._state_value.set_value(state_str, v)
    # Update the cumulative value of the state.
    self._cumulative_state_value.add_value(state_str, learning_rate * v)
    return v

  def get_projected_policy(self) -> policy_lib.Policy:
    """Returns the projected policy."""
    return ProjectedPolicy(self._game, list(range(self._game.num_players())),
                           self._cumulative_state_value)

  def iteration(self, learning_rate: Optional[float] = None):
    """An iteration of Mirror Descent."""
    self._md_step += 1
    # TODO(sertan): Fix me.
    self._state_value = value.TabularValueFunction(self._game)
    for state in self._root_states:
      self.eval_state(state, learning_rate if learning_rate else self._lr)
    self._policy = self.get_projected_policy()
    self._distribution = distribution.DistributionPolicy(
        self._game, self._policy)

  def get_policy(self) -> policy_lib.Policy:
    return self._policy

  @property
  def distribution(self) -> distribution.DistributionPolicy:
    return self._distribution



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fixed Point."""

from typing import Optional

from open_spiel.python import policy as policy_lib
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import greedy_policy
from open_spiel.python.mfg.algorithms import softmax_policy
import pyspiel


class FixedPoint(object):
  """The fixed point algorithm.

  This algorithm is based on Banach-Picard iterations for the fixed point
  operator characterizing the Nash equilibrium. At each iteration, the policy is
  updated by computing a best response against the current mean-field or a
  regularized version that is obtained by taking a softmax with respect to the
  optimal Q-function, and the mean-field is updated by taking the mean-field
  induced by the current policy.
  """

  def __init__(self, game: pyspiel.Game, temperature: Optional[float] = None):
    """Initializes the algorithm.

    Args:
      game: The game to analyze.
      temperature: If set, then instead of the greedy policy a softmax policy
        with the specified temperature will be used to update the policy at each
        iteration.
    """
    self._game = game
    self._temperature = temperature
    self._policy = policy_lib.UniformRandomPolicy(self._game)
    self._distribution = distribution.DistributionPolicy(game, self._policy)

  def iteration(self):
    """An itertion of Fixed Point."""
    # Calculate the current distribution and the best response.
    distrib = distribution.DistributionPolicy(self._game, self._policy)
    br_value = best_response_value.BestResponse(
        self._game, distrib, value.TabularValueFunction(self._game))

    # Policy is either greedy or softmax with respect to the best response if
    # temperature is specified.
    player_ids = list(range(self._game.num_players()))
    if self._temperature is None:
      self._policy = greedy_policy.GreedyPolicy(self._game, player_ids,
                                                br_value)
    else:
      self._policy = softmax_policy.SoftmaxPolicy(self._game, player_ids,
                                                  self._temperature, br_value)

    self._distribution = distribution.DistributionPolicy(
        self._game, self._policy)

  def get_policy(self) -> policy_lib.Policy:
    return self._policy

  @property
  def distribution(self) -> distribution.DistributionPolicy:
    return self._distribution



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test Python bindings for game transforms."""

from absl.testing import absltest

import pyspiel


class RepeatedGameTest(absltest.TestCase):

  def test_create_repeated_game(self):
    """Test both create_repeated_game function signatures."""
    repeated_game = pyspiel.create_repeated_game("matrix_rps",
                                                 {"num_repetitions": 10})
    assert repeated_game.utility_sum() == 0
    state = repeated_game.new_initial_state()
    for _ in range(10):
      state.apply_actions([0, 0])
    assert state.is_terminal()

    stage_game = pyspiel.load_game("matrix_mp")
    repeated_game = pyspiel.create_repeated_game(stage_game,
                                                 {"num_repetitions": 5})
    state = repeated_game.new_initial_state()
    for _ in range(5):
      state.apply_actions([0, 0])
    assert state.is_terminal()

    stage_game = pyspiel.load_game("matrix_pd")
    repeated_game = pyspiel.create_repeated_game(stage_game,
                                                 {"num_repetitions": 5})
    assert repeated_game.utility_sum() is None


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for GamutGenerator."""

from absl import app
from absl.testing import absltest
from absl.testing import parameterized
from open_spiel.python.egt.utils import game_payoffs_array
import pyspiel


class GamutGeneratorTest(parameterized.TestCase):

  def _gamut_generator(self):
    return pyspiel.GamutGenerator(
        "gamut.jar"
    )

  @parameterized.parameters(
      "-g BertrandOligopoly -players 2 -actions 4 -random_params",
      "-g UniformLEG-CG -players 2 -actions 4 -random_params",
      "-g PolymatrixGame-SW -players 2 -actions 4 -random_params",
      "-g GraphicalGame-SW -players 2 -actions 4 -random_params",
      "-g BidirectionalLEG-CG -players 2 -actions 4 -random_params",
      "-g CovariantGame -players 2 -actions 4 -random_params",
      "-g DispersionGame -players 2 -actions 4 -random_params",
      "-g MinimumEffortGame -players 2 -actions 4 -random_params",
      "-g RandomGame -players 2 -actions 4 -random_params",
      "-g TravelersDilemma -players 2 -actions 4 -random_params",
  )
  def test_generate_game(self, game_str):
    generator = self._gamut_generator()
    # Using a string of arguments.
    game = generator.generate_game(game_str)
    self.assertIsNotNone(game)

    payoff_tensor = game_payoffs_array(game)
    self.assertEqual(payoff_tensor.shape, (2, 4, 4))

  def test_gamut_api(self):
    generator = self._gamut_generator()

    # See the documentation at http://gamut.stanford.edu/ for the commands
    # needed to generate the various different games.

    # Using a string of arguments.
    game = generator.generate_game(
        "-g RandomGame -players 4 -normalize -min_payoff 0 "
        + "-max_payoff 150 -actions 2 4 5 7"
    )
    self.assertIsNotNone(game)

    # Using a list of arguments.
    game = generator.generate_game([
        "-g",
        "RandomGame",
        "-players",
        "4",
        "-normalize",
        "-min_payoff",
        "0",
        "-max_payoff",
        "150",
        "-actions",
        "2",
        "4",
        "5",
        "7",
    ])
    self.assertIsNotNone(game)

    # Using a list of arguments.
    matrix_game = generator.generate_matrix_game([
        "-g",
        "RandomGame",
        "-players",
        "2",
        "-normalize",
        "-min_payoff",
        "0",
        "-max_payoff",
        "150",
        "-actions",
        "10",
        "15",
    ])
    self.assertIsNotNone(matrix_game)
    print(matrix_game.new_initial_state())
    payoff_matrix = game_payoffs_array(matrix_game)
    print(payoff_matrix.shape)
    print(payoff_matrix)

    # Using a list of arguments.
    tensor_game = generator.generate_game([
        "-g",
        "RandomGame",
        "-players",
        "4",
        "-normalize",
        "-min_payoff",
        "0",
        "-max_payoff",
        "150",
        "-actions",
        "2",
        "4",
        "5",
        "7",
    ])
    self.assertIsNotNone(tensor_game)
    payoff_tensor = game_payoffs_array(tensor_game)
    print(payoff_tensor.shape)


def main(_):
  absltest.main()


if __name__ == "__main__":
  # Calling main via app.run here is necessary for internal uses.
  app.run(main)
